from o3seespy.command.nd_material.base_material import NDMaterialBase
from o3seespy.command.common import update_material_stage


class PM4Sand(NDMaterialBase):
    op_type = "PM4Sand"

    def __init__(self, osi, d_r, g_o, h_po, den, p_atm, h_o=-1.0, e_max=0.8, e_min=0.5, n_b=0.5, n_d=0.1, a_do=None,
                 z_max=None, c_z=250.0, c_e=None, phi_cv=33.0, nu=0.3, g_degr=2.0, c_dr=None, c_kaf=None, q_bolt=10.0,
                 r_bolt=1.5, m_par=0.01, f_sed=None, p_sed=None):
        """
        This command is used to construct a 2-dimensional PM4Sand material.

        Parameters
        ==========
        d_r: float
            Relative density, in fraction
        g_o: float
            Shear modulus constant
        h_po: float
            Contraction rate parameter
        den: float
            Mass density of the material
        p_atm: float, optional
            Atmospheric pressure
        h_o: float, optional
            Variable that adjusts the ratio of plastic modulus to elastic modulus
        e_max: float, optional
            Maximum and minimum void ratios
        e_min: float, optional
            Maximum and minimum void ratios
        n_b: float, optional
            Bounding surface parameter, :math:`nb \ge 0`
        n_d: float, optional
            Dilatancy surface parameter :math:`nd \ge 0`
        a_do: float, optional
            Dilatancy parameter, will be computed at the time of initialization if input value is negative
        z_max: float, optional
            Fabric-dilatancy tensor parameter
        c_z: float, optional
            Fabric-dilatancy tensor parameter
        c_e: float, optional
            Variable that adjusts the rate of strain accumulation in cyclic loading
        phi_cv: float, optional
            Critical state effective friction angle
        nu: float, optional
            Poisson's ratio
        c_gd: float, optional
            Variable that adjusts degradation of elastic modulus with accumulation of fabric
        c_dr: float,   optional
            Variable that controls the rotated dilatancy surface
        c_kaf: float,  optional
            Variable that controls the effect that sustained
                                          con shear stresses have on plastic modulus
        q_bolt: float,     optional
            Critical state line parameter
        r_bolt: float,     optional
            Critical state line parameter
        m_par: float,     optional
            Yield surface constant (radius of yield surface in stress ratio space)
        f_sed: float, optional,
            Variable that controls the minimum value the reduction factor of the elastic moduli
            can get during reconsolidation
        p_sed: float, optional
            Mean effective stress up to which reconsolidation strains are enhanced
        """
        self.osi = osi
        self.d_r = float(d_r)
        self.g_o = float(g_o)
        self.h_po = float(h_po)
        self.den = float(den)
        self.p_atm = float(p_atm)
        self.h_o = float(h_o)
        self.e_max = float(e_max)
        self.e_min = float(e_min)
        self.n_b = float(n_b)
        self.n_d = float(n_d)
        if a_do is None:
            self.a_do = -1.0
        else:
            self.a_do = float(a_do)
        if z_max is None:
            self.z_max = -1.0
        else:
            self.z_max = float(z_max)
        self.c_z = float(c_z)
        if c_e is None:
            self.c_e = -1.0
        else:
            self.c_e = float(c_e)
        self.phi_cv = float(phi_cv)
        self.nu = float(nu)
        self.g_degr = float(g_degr)
        if c_dr is None:
            self.c_dr = -1.0
        else:
            self.c_dr = float(c_dr)
        if c_kaf is None:
            self.c_kaf = -1.0
        else:
            self.c_kaf = float(c_kaf)
        self.q_bolt = float(q_bolt)
        self.r_bolt = float(r_bolt)
        self.m_par = float(m_par)
        if f_sed is None:
            self.f_sed = -1.0
        else:
            self.f_sed = float(f_sed)
        if p_sed is None:
            self.p_sed = -1.0
        else:
            self.p_sed = float(p_sed)

        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat

        self._parameters = [self.op_type, self._tag, self.d_r, self.g_o, self.h_po, self.den, self.p_atm, self.h_o,
                            self.e_max, self.e_min, self.n_b, self.n_d, self.a_do, self.z_max, self.c_z, self.c_e, self.phi_cv,
                            self.nu, self.g_degr, self.c_dr, self.c_kaf, self.q_bolt, self.r_bolt, self.m_par, self.f_sed,
                            self.p_sed]

        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)

    def pre_dynamic(self, osi):  # deprecated
        update_material_stage(osi, self, stage=1)
        # opw.set_parameter(osi, value=0, eles=[ele], args=['FirstCall', 1])

    def update_to_nonlinear(self):
        from o3seespy import update_material_stage
        update_material_stage(self.osi, self, 1)

    def set_nu(self, nu, ele=None, eles=None):
        from o3seespy import set_parameter
        if ele is not None:
            set_parameter(self.osi, value=nu, eles=[ele], args=['poissonRatio', 1])
        if eles is not None:
            set_parameter(self.osi, value=nu, eles=eles, args=['poissonRatio', 1])

    def set_material_state(self, value, ele=None, eles=None):
        from o3seespy import set_parameter
        if ele is not None:
            set_parameter(self.osi, value=value, eles=[ele], args=['materialState', 1])
        if eles is not None:
            set_parameter(self.osi, value=value, eles=eles, args=['materialState', 1])

    def set_first_call(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, pstr='FirstCall', pval=self.tag, value=value, ele=ele, eles=eles)








import o3seespy as o3  # for testing only
import pytest


def test_hardening():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.Hardening(osi, e_mod=1.0, sigma_y=1.0, h_iso=1.0, h_kin=1.0, eta=0.0)


@pytest.mark.skip()
def test_cast():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.Cast(osi, n=1, bo=1.0, h=1.0, fy=1.0, e_mod=1.0, big_l=1.0, b=1.0, ro=1.0, c_r1=1.0, c_r2=1.0, a1=None, a2=1.0, a3=None, a4=1.0)


def test_viscous_damper():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.ViscousDamper(osi, k_el=1.0, cd=1.0, alpha=1.0, l_gap=0.0, nm=1, rel_tol=1e-6, abs_tol=1e-10, max_half=15)


def test_bilinear_oil_damper():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.BilinearOilDamper(osi, k_el=1.0, cd=1.0, fr=1.0, p=1.0, l_gap=0.0, nm=1, rel_tol=1e-6, abs_tol=1e-10, max_half=15)


def test_bilin():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.Bilin(osi, k0=1.0, as__plus=1.0, as__neg=1.0, my__plus=1.0, my__neg=1.0, lamda_s=1.0, lamda_c=1.0, lamda_a=1.0, lamda_k=1.0, c_s=1.0, c_c=1.0, c_a=1.0, c_k=1.0, theta_p__plus=1.0, theta_p__neg=1.0, theta_pc__plus=1.0, theta_pc__neg=1.0, res__pos=1.0, res__neg=1.0, theta_u__plus=1.0, theta_u__neg=1.0, d__plus=1.0, d__neg=1.0, n_factor=0.0)


def test_mod_imk_peak_oriented():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.ModIMKPeakOriented(osi, k0=1.0, as__plus=1.0, as__neg=1.0, my__plus=1.0, my__neg=1.0, lamda_s=1.0, lamda_c=1.0, lamda_a=1.0, lamda_k=1.0, c_s=1.0, c_c=1.0, c_a=1.0, c_k=1.0, theta_p__plus=1.0, theta_p__neg=1.0, theta_pc__plus=1.0, theta_pc__neg=1.0, res__pos=1.0, res__neg=1.0, theta_u__plus=1.0, theta_u__neg=1.0, d__plus=1.0, d__neg=1.0)


def test_mod_imk_pinching():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.ModIMKPinching(osi, k0=1.0, as__plus=1.0, as__neg=1.0, my__plus=1.0, my__neg=1.0, fpr_pos=1.0, fpr_neg=1.0, a_pinch=1.0, lamda_s=1.0, lamda_c=1.0, lamda_a=1.0, lamda_k=1.0, c_s=1.0, c_c=1.0, c_a=1.0, c_k=1.0, theta_p__plus=1.0, theta_p__neg=1.0, theta_pc__plus=1.0, theta_pc__neg=1.0, res__pos=1.0, res__neg=1.0, theta_u__plus=1.0, theta_u__neg=1.0, d__plus=1.0, d__neg=1.0)


def test_saws():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.SAWS(osi, f0=1.0, fi=1.0, du=1.0, s0=1.0, r1=1.0, r2=1.0, r3=1.0, r4=1.0, alpha=1.0, beta=1.0)


@pytest.mark.skip()
def test_bar_slip():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.BarSlip(osi, fc=1.0, fy=1.0, es=1.0, fu=1.0, eh=1.0, db=1.0, ld=1.0, nb=1, depth=1.0, height=1.0, anc_lratio=1.0, bs_flag="Strong", otype="beamtop", damage='Damage', unit='psi')


def test_bond_sp01():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.BondSP01(osi, fy=1.0, sy=1.0, fu=1.0, su=1.0, b=1.0, big_r=1.0)


def test_fatigue():
    osi = o3.OpenSeesInstance(ndm=2)
    other = o3.uniaxial_material.Hardening(osi, e_mod=1.0, sigma_y=1.0, h_iso=1.0, h_kin=1.0, eta=0.0)
    o3.uniaxial_material.Fatigue(osi, other=other, e0=0.191, m=-0.458, min=-1e16, max=1e16)


@pytest.mark.skip()
def test_impact_material():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.ImpactMaterial(osi, k1=1.0, k2=1.0, sigy=1.0, gap=1.0)


@pytest.mark.skip()
def test_hyperbolic_gap_material():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.HyperbolicGapMaterial(osi, kmax=1.0, kur=2.0, rf=1.0, fult=2.0, gap=0.0)


@pytest.mark.skip()
def test_limit_state():
    osi = o3.OpenSeesInstance(ndm=2)
    curve = 1
    o3.uniaxial_material.LimitState(osi, s1p=1.0, e1p=1.0, s2p=1.0, e2p=1.0, s3p=1.0, e3p=1.0, s1n=1.0, e1n=1.0, s2n=1.0, e2n=1.0, s3n=1.0, e3n=1.0, pinch_x=1.0, pinch_y=1.0, damage1=1.0, damage2=1.0, beta=1.0, curve=curve, curve_type=1)


def test_min_max():
    osi = o3.OpenSeesInstance(ndm=2)
    other = o3.uniaxial_material.Hardening(osi, e_mod=1.0, sigma_y=1.0, h_iso=1.0, h_kin=1.0, eta=0.0)
    o3.uniaxial_material.MinMax(osi, other=other, min_strain=1e-16, max_strain=1e16)


def test_elastic_bilin():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.ElasticBilin(osi, ep1=1.0, ep2=1.0, eps_p2=1.0, en1=None, en2=None, eps_n2=None)


def test_elastic_multi_linear():
    osi = o3.OpenSeesInstance(ndm=2)
    strain = [1.0, 1.0]
    stress = [1.0, 1.0]
    o3.uniaxial_material.ElasticMultiLinear(osi, eta=0.0, strain=strain, stress=stress)


def test_multi_linear():
    osi = o3.OpenSeesInstance(ndm=2)
    pts = [0.01, 1.0, 0.02, 2.0]
    o3.uniaxial_material.MultiLinear(osi, pts=pts)


def test_init_strain_material():
    osi = o3.OpenSeesInstance(ndm=2)
    other = o3.uniaxial_material.ModIMKPinching(osi, k0=1.0, as__plus=1.0, as__neg=1.0, my__plus=1.0, my__neg=1.0, fpr_pos=1.0, fpr_neg=1.0, a_pinch=1.0, lamda_s=1.0, lamda_c=1.0, lamda_a=1.0, lamda_k=1.0, c_s=1.0, c_c=1.0, c_a=1.0, c_k=1.0, theta_p__plus=1.0, theta_p__neg=1.0, theta_pc__plus=1.0, theta_pc__neg=1.0, res__pos=1.0, res__neg=1.0, theta_u__plus=1.0, theta_u__neg=1.0, d__plus=1.0, d__neg=1.0)
    o3.uniaxial_material.InitStrainMaterial(osi, other=other, init_strain=1.0)


def test_init_stress_material():
    osi = o3.OpenSeesInstance(ndm=2)
    other = o3.uniaxial_material.ModIMKPinching(osi, k0=1.0, as__plus=1.0, as__neg=1.0, my__plus=1.0, my__neg=1.0,
                                                fpr_pos=1.0, fpr_neg=1.0, a_pinch=1.0, lamda_s=1.0, lamda_c=1.0,
                                                lamda_a=1.0, lamda_k=1.0, c_s=1.0, c_c=1.0, c_a=1.0, c_k=1.0,
                                                theta_p__plus=1.0, theta_p__neg=1.0, theta_pc__plus=1.0,
                                                theta_pc__neg=1.0, res__pos=1.0, res__neg=1.0, theta_u__plus=1.0,
                                                theta_u__neg=1.0, d__plus=1.0, d__neg=1.0)
    o3.uniaxial_material.InitStressMaterial(osi, other=other, init_stress=-1.0)


def test_path_independent():
    osi = o3.OpenSeesInstance(ndm=2)
    other = o3.uniaxial_material.Hardening(osi, e_mod=1.0, sigma_y=1.0, h_iso=1.0, h_kin=1.0, eta=0.0)
    o3.uniaxial_material.PathIndependent(osi, other=other)


def test_ecc01():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.ECC01(osi, sigt0=1.0, epst0=1.0, sigt1=1.0, epst1=1.0, epst2=1.0, sigc0=1.0, epsc0=1.0, epsc1=1.0, alpha_t1=1.0, alpha_t2=1.0, alpha_c=1.0, alpha_cu=1.0, beta_t=1.0, beta_c=1.0)


def test_self_centering():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.SelfCentering(osi, k1=1.0, k2=1.0, sig_act=1.0, beta=1.0, eps_slip=0, eps_bear=0, r_bear=None)


def test_viscous():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.Viscous(osi, big_c=1.0, alpha=1.0)


def test_bouc_wen():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.BoucWen(osi, alpha=1.0, ko=1.0, n=1.0, gamma=1.0, beta=1.0, ao=1.0, delta_a=1.0, delta_nu=1.0, delta_eta=1.0)


def test_bwbn():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.BWBN(osi, alpha=1.0, ko=1.0, n=1.0, gamma=1.0, beta=1.0, ao=1.0, q=1.0, zetas=1.0, p=1.0, shi=1.0, delta_shi=1.0, lamb=1.0, tol=1.0, max_iter=1.0)


def test_axial_sp():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.AxialSp(osi, sce=1.0, fty=1.0, fcy=1.0, bte=1.0, bty=1.0, bcy=1.0, fcr=1.0)


def test_axial_sp_hd():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.AxialSpHD(osi, sce=1.0, fty=1.0, fcy=1.0, bte=1.0, bty=1.0, bth=1.0, bcy=1.0, fcr=1.0, ath=1.0)


@pytest.mark.skip()
def test_cfswswp():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.CFSWSWP(osi, height=1.0, width=1.0, fut=1.0, tf=1.0, ife=1.0, ifi=1.0, ts=1.0, np=1.0, ds=1.0, vs=1.0, sc=1.0, nc=1.0, otype=1, opening_area=1.0, opening_length=1.0)


def test_cfssswp():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.CFSSSWP(osi, height=1.0, width=1.0, fuf=1.0, fyf=1.0, tf=1.0, af=1.0, fus=1.0, fys=1.0, ts=1.0, np=1.0, ds=1.0, vs=1.0, sc=1.0, dt=1.0, opening_area=1.0, opening_length=1.0)


if __name__ == '__main__':
    test_init_stress_material()


import o3seespy as o3


def load_element(nu_init, nu_setp):
    esig_v0 = 100.0

    osi = o3.OpenSeesInstance(ndm=2, ndf=2)

    mat = o3.nd_material.ElasticIsotropic(osi, 1.0e6, nu=nu_init)
    h_ele = 1.
    nodes = [
        o3.node.Node(osi, 0.0, 0.0),
        o3.node.Node(osi, h_ele, 0.0),
        o3.node.Node(osi, h_ele, h_ele),
        o3.node.Node(osi, 0.0, h_ele)
    ]

    # Fix bottom node
    o3.Fix2DOF(osi, nodes[0], o3.cc.FIXED, o3.cc.FIXED)
    o3.Fix2DOF(osi, nodes[1], o3.cc.FIXED, o3.cc.FIXED)
    o3.Fix2DOF(osi, nodes[2], o3.cc.FREE, o3.cc.FREE)
    o3.Fix2DOF(osi, nodes[3], o3.cc.FREE, o3.cc.FREE)
    # Set out-of-plane DOFs to be slaved
    o3.EqualDOF(osi, nodes[2], nodes[3], [o3.cc.X, o3.cc.Y])

    ele = o3.element.SSPquad(osi, nodes, mat, 'PlaneStrain', 1, 0.0, 0.0)

    o3.constraints.Transformation(osi)
    o3.test_check.NormDispIncr(osi, tol=1.0e-3, max_iter=35, p_flag=0)
    o3.algorithm.Newton(osi)
    o3.numberer.RCM(osi)
    o3.system.FullGeneral(osi)
    o3.integrator.DisplacementControl(osi, nodes[2], o3.cc.DOF2D_Y, 0.005)
    # o3.rayleigh.Rayleigh(osi, a0, a1, 0.0, 0.0)
    o3.analysis.Static(osi)
    o3.set_parameter(osi, value=nu_setp, eles=[ele], args=['nu', 1])

    # Add static vertical pressure and stress bias
    # time_series = o3.time_series.Path(osi, time=[0, 100, 1e10], values=[0, 1, 1])
    # o3.pattern.Plain(osi, time_series)
    ts0 = o3.time_series.Linear(osi, factor=1)
    o3.pattern.Plain(osi, ts0)
    o3.Load(osi, nodes[2], [0, -esig_v0 / 2])
    o3.Load(osi, nodes[3], [0, -esig_v0 / 2])

    o3.analyze(osi, num_inc=100)
    stresses = o3.get_ele_response(osi, ele, 'stress')
    print('init_stress0: ', stresses)






def get_pid(osi):
    """
    Get the processor ID of the calling processor.

    Parameters
    ----------
    osi: o3seespy.OpenSeesInstance
    """
    _parameters = []
    return osi.to_process("getPID", _parameters)


def get_np(osi):
    """
    Get total number of processors.

    Parameters
    ----------
    osi: o3seespy.OpenSeesInstance
    """
    _parameters = []
    return osi.to_process("getNP", _parameters)


def barrier(osi):
    """
    Set a barrier for all processors, i.e.,faster processors will pause here to wait for all processorsto reach to this
    point.

    Parameters
    ----------
    osi: o3seespy.OpenSeesInstance
    """
    _parameters = []
    return osi.to_process("barrier", _parameters)


def send(osi, pid, data):
    """
    Send information to another processor.

    Parameters
    ----------
    osi: o3seespy.OpenSeesInstance
    pid: int
            Id of processor where data is sent to
    data: str
            Can be a string
    """
    pid = int(pid)
    _parameters = [pid, *data]
    return osi.to_process("send", _parameters)


def recv(osi, pid):
    """
    Receive information from another processor.

    Parameters
    ----------
    osi: o3seespy.OpenSeesInstance
    pid: str
            If ``pid`` is ``'any'``, the processor can receive data from any processor.
    """
    _parameters = [pid]
    return osi.to_process("recv", _parameters)


def bcast(osi):
    """
    Broadcast information from processor 0 to all processors.

    Parameters
    ----------
    osi: o3seespy.OpenSeesInstance
    """
    _parameters = []
    return osi.to_process("Bcast", _parameters)


def set_start_node_tag(osi):
    """
    Set the starting node tag for the :doc:`mesh`.The purpose of this command is to controlthe node tags generated by
    the :doc:`mesh`. Somenodes are shared by processors, which musthave same tags. Nodes which are unique toa processor
    must have uniques tags acrossall processors.

    Parameters
    ----------
    osi: o3seespy.OpenSeesInstance
    """
    _parameters = []
    return osi.to_process("setStartNodeTag", _parameters)


def domain_change(osi):
    """
    Mark the domain has changed manually.This is used to notify processors whose domain is not changed,but the domain in
    other processors have changed.

    Parameters
    ----------
    osi: o3seespy.OpenSeesInstance
    """
    _parameters = []
    return osi.to_process("domainChange", _parameters)


def partition(osi, ncuts, niters: int=None, ufactor: int=None, info=False):
    """
    In a parallel environment, this command partitions the model. It requires that all processorshave the exact same
    model to be partitioned. 

    Parameters
    ----------
    osi: o3seespy.OpenSeesInstance
    ncuts: int
            Specifies the number of different partitionings that it will compute.  the final partitioning is the one
            that achieves the best edge cut or communication volume.  (optional default is 1).
    niters: int, optional
            Specifies the number of iterations for the refinement algorithms  at each sobjecte of the uncoarsening
            process.  (optional default is 10).
    ufactor: int, optional
            Specifies the maximum allowed load imbalance among the partitions.  (optional default is 30, indicating a
            load imbalance of 1.03).
    info: bool
            Print information. 
    """
    ncuts = int(ncuts)
    if niters is not None:
        niters = int(niters)
    if ufactor is not None:
        ufactor = int(ufactor)
    _parameters = [ncuts]
    if niters is not None:
        _parameters += ['-niter', niters]
    if ufactor is not None:
        _parameters += ['-ufactor', ufactor]
    if info:
        _parameters += ['-info']
    return osi.to_process("partition", _parameters)



import o3seespy as o3  # for testing only


def test_ss_pquad_up():
    osi = o3.OpenSeesInstance(ndm=2)
    obj = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
    ele_nodes = [o3.node.Node(osi, 0.0, x) for x in range(4)]
    o3.element.SSPquadUP(osi, ele_nodes=ele_nodes, mat=obj, thick=1.0, f_bulk=1.0, f_den=1.0, k1=1.0, k2=1.0, void=1.0, alpha=1.0, b1=0.0, b2=0.0)


def test_ss_pbrick_up():
    osi = o3.OpenSeesInstance(ndm=2)
    obj = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
    ele_nodes = [o3.node.Node(osi, 0.0, x) for x in range(8)]
    o3.element.SSPbrickUP(osi, ele_nodes=ele_nodes, mat=obj, f_bulk=1.0, f_den=1.0, k1=1.0, k2=1.0, k3=1.0, void=0.5, alpha=1.0e-5, b1=1.0, b2=1.0, b3=1.0)


if __name__ == '__main__':
    test_ss_pquad_up()



from o3seespy.command.element.base_element import ElementBase


class SSPquadUP(ElementBase):
    """
    The SSPquadUP Element Class
    
    This command is used to construct a SSPquadUP element object.

    
    """
    op_type = 'SSPquadUP'

    def __init__(self, osi, ele_nodes, mat, thick, f_bulk, f_den, k1, k2, void, alpha, b1=0.0, b2=0.0):
        """
        Initial method for SSPquadUP

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of four element nodes in counter-clockwise order
        mat: obj
            Unique integer object associated with previously-defined ndmaterial object
        thick: float
            Thickness of the element in out-of-plane direction
        f_bulk: float
            Bulk modulus of the pore fluid
        f_den: float
            Mass density of the pore fluid
        k1: float
            Permeability coefficients in global x- and y-directions, respectively
        k2: float
            Permeability coefficients in global x- and y-directions, respectively
        void: float
            Voids ratio
        alpha: float
            Spatial pressure field stabilization parameter (see discussion below for more information)
        b1: float, optional
            Constant body forces in global x- and y-directions, respectively (optional, default = 0.0) - see note 3
        b2: float, optional
            Constant body forces in global x- and y-directions, respectively (optional, default = 0.0) - see note 3

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> obj = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
        >>> ele_nodes = [o3.node.Node(osi, 0.0, x) for x in range(4)]
        >>> o3.element.SSPquadUP(osi, ele_nodes=ele_nodes, mat=obj, thick=1.0, f_bulk=1.0, f_den=1.0, k1=1.0, k2=1.0, void=1.0, alpha=1.0, b1=0.0, b2=0.0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.mat = mat
        self.thick = float(thick)
        self.f_bulk = float(f_bulk)
        self.f_den = float(f_den)
        self.k1 = float(k1)
        self.k2 = float(k2)
        self.void = float(void)
        self.alpha = float(alpha)
        self.b1 = float(b1)
        self.b2 = float(b2)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.mat.tag, self.thick, self.f_bulk, self.f_den, self.k1, self.k2, self.void, self.alpha, self.b1, self.b2]
        self.to_process(osi)


class SSPbrickUP(ElementBase):
    """
    The SSPbrickUP Element Class
    
    This command is used to construct a SSPbrickUP element object.

    
    """
    op_type = 'SSPbrickUP'

    def __init__(self, osi, ele_nodes, mat, f_bulk, f_den, k1, k2, k3, void, alpha, b1, b2, b3):
        """
        Initial method for SSPbrickUP

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of eight element nodes in counter-clockwise order
        mat: obj
            Unique integer object associated with previously-defined ndmaterial object
        f_bulk: float
            Bulk modulus of the pore fluid
        f_den: float
            Mass density of the pore fluid
        k1: float
            Permeability coefficients in global x-, y-, and z-directions, respectively
        k2: float
            Permeability coefficients in global x-, y-, and z-directions, respectively
        k3: float
            Permeability coefficients in global x-, y-, and z-directions, respectively
        void: float
            Voids ratio
        alpha: float
            Spatial pressure field stabilization parameter (see discussion below for more information)
        b1: float
            Constant body forces in global x-, y-, and z-directions, respectively (optional, default = 0.0) - see note 3
        b2: float
            Constant body forces in global x-, y-, and z-directions, respectively (optional, default = 0.0) - see note 3
        b3: float
            Constant body forces in global x-, y-, and z-directions, respectively (optional, default = 0.0) - see note 3

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> obj = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
        >>> ele_nodes = [o3.node.Node(osi, 0.0, x) for x in range(8)]
        >>> o3.element.SSPbrickUP(osi, ele_nodes=ele_nodes, mat=obj, f_bulk=1.0, f_den=1.0, k1=1.0, k2=1.0, k3=1.0, void=0.5, alpha=1.0e-5, b1=1.0, b2=1.0, b3=1.0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.mat = mat
        self.f_bulk = float(f_bulk)
        self.f_den = float(f_den)
        self.k1 = float(k1)
        self.k2 = float(k2)
        self.k3 = float(k3)
        self.void = float(void)
        self.alpha = float(alpha)
        self.b1 = float(b1)
        self.b2 = float(b2)
        self.b3 = float(b3)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.mat.tag, self.f_bulk, self.f_den, self.k1, self.k2, self.k3, self.void, self.alpha, self.b1, self.b2, self.b3]
        self.to_process(osi)



import o3seespy as o3


def test_fix1dof():
    osi = o3.OpenSeesInstance(ndm=2, ndf=1)
    node = o3.node.Node(osi, 3, 4)
    o3.Fix1DOF(osi, node, o3.cc.FIXED)
    sn = o3.node.build_regular_node_mesh(osi, [3, 4], [1, 6])
    o3.Fix1DOFMulti(osi, sn[0], o3.cc.FREE)


def test_fix2dof():
    osi = o3.OpenSeesInstance(ndm=2, ndf=2)
    node = o3.node.Node(osi, 3, 4)
    o3.Fix2DOF(osi, node, o3.cc.FIXED, o3.cc.FIXED)
    sn = o3.node.build_regular_node_mesh(osi, [3, 4], [1, 6])
    o3.Fix2DOFMulti(osi, sn[0], o3.cc.FREE, o3.cc.FIXED)


def test_fix3dof():
    osi = o3.OpenSeesInstance(ndm=2, ndf=3)
    node = o3.node.Node(osi, 3, 4)
    o3.Fix3DOF(osi, node, o3.cc.FIXED, o3.cc.FIXED, o3.cc.FIXED)
    sn = o3.node.build_regular_node_mesh(osi, [3, 4], [1, 6])
    o3.Fix3DOFMulti(osi, sn[0], o3.cc.FREE, o3.cc.FIXED, o3.cc.FIXED)


def test_fix6dof():
    osi = o3.OpenSeesInstance(ndm=2, ndf=6)
    node = o3.node.Node(osi, 3, 4)
    o3.Fix6DOF(osi, node, o3.cc.FIXED, o3.cc.FIXED, o3.cc.FIXED, o3.cc.FIXED, o3.cc.FIXED, o3.cc.FIXED)
    sn = o3.node.build_regular_node_mesh(osi, [3, 4], [1, 6])
    o3.Fix6DOFMulti(osi, sn[0], o3.cc.FREE, o3.cc.FIXED, o3.cc.FIXED, o3.cc.FIXED, o3.cc.FIXED, o3.cc.FIXED)


def test_equal_dof():
    osi = o3.OpenSeesInstance(ndm=2, ndf=2)
    node1 = o3.node.Node(osi, 3, 4)
    node2 = o3.node.Node(osi, 3, 4)
    node3 = o3.node.Node(osi, 3, 4)
    o3.EqualDOF(osi, node1, node2, [o3.cc.DOF2D_X, o3.cc.DOF2D_Y])  # 1-to-1
    sn = o3.node.build_regular_node_mesh(osi, [3, 4, 5], [1, 6])
    o3.EqualDOFMulti(osi, node3, sn[0], [o3.cc.DOF2D_X, o3.cc.DOF2D_Y])  # 1-to-many
    o3.EqualDOFMulti(osi, sn[1], sn[2], [o3.cc.DOF2D_Y])  # many 1-to-1s


def test_add_fixity_to_dof():
    osi = o3.OpenSeesInstance(ndm=3, ndf=3)

    n1 = o3.node.Node(osi, 1., 1., 1.)
    o3.Fix3DOF(osi, n1, o3.cc.FIXED, o3.cc.FREE, o3.cc.FIXED)
    o3.Fix3DOF(osi, n1, o3.cc.FREE, o3.cc.FREE, o3.cc.FREE)
    # o3.Fix3DOF(osi, n1, o3.cc.FREE, o3.cc.FREE, o3.cc.FIXED)  # fails with this
    o3.add_fixity_to_dof(osi, o3.cc.DOF2D_ROTZ, [n1])





import o3seespy as o3  # for testing only
import pytest
import numpy as np


def test_ele_load_uniform():
    osi = o3.OpenSeesInstance(ndm=2, state=3)
    ele_len = 2.0
    coords = [[0, 0], [ele_len, 0]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    transf = o3.geom_transf.Linear2D(osi, [])
    ele = o3.element.ElasticBeamColumn2D(osi, ele_nodes=ele_nodes,
                                   area=1.0, e_mod=1.0, iz=1.0, transf=transf, mass=1.0, c_mass="string")

    for i, node in enumerate(ele_nodes):
        if i == 0:
            o3.Fix3DOF(osi, node, o3.cc.FIXED, o3.cc.FIXED, o3.cc.FIXED)
        else:
            o3.Fix3DOF(osi, node, o3.cc.FREE, o3.cc.FIXED, o3.cc.FIXED)

    ts_po = o3.time_series.Linear(osi, factor=1)
    o3.pattern.Plain(osi, ts_po)
    udl = 10.
    o3.EleLoad2DUniform(osi, ele, w_y=-udl)

    tol = 1.0e-4
    o3.constraints.Plain(osi)
    o3.numberer.RCM(osi)
    o3.system.BandGen(osi)
    o3.test_check.NormDispIncr(osi, tol, 6)
    o3.algorithm.Newton(osi)
    n_steps_gravity = 1
    d_gravity = 1. / n_steps_gravity
    o3.integrator.LoadControl(osi, d_gravity, num_iter=10)
    o3.analysis.Static(osi)
    o3.analyze(osi, n_steps_gravity)
    o3.gen_reactions(osi)
    ele_loads = o3.get_ele_response(osi, ele, 'force')
    assert np.isclose(ele_loads[0], 0.0)
    assert np.isclose(ele_loads[1], udl * ele_len / 2)
    assert np.isclose(ele_loads[2], udl * ele_len ** 2 / 12)
    assert np.isclose(ele_loads[3], 0.0)
    assert np.isclose(ele_loads[4], udl * ele_len / 2)
    assert np.isclose(ele_loads[5], -udl * ele_len ** 2 / 12)
    load = o3.get_node_reaction(osi, ele_nodes[0], o3.cc.DOF_Y)
    assert np.isclose(load, udl * ele_len / 2)


if __name__ == '__main__':
    test_ele_load_uniform()



from o3seespy.command.element.base_element import ElementBase


class Tri31(ElementBase):
    """
    The Tri31 Element Class
    
    This command is used to construct a constant strain triangular element (Tri31) which uses three nodes and one
    integration points.

    
    """
    op_type = 'Tri31'

    def __init__(self, osi, ele_nodes, thick, otype, mat, pressure, rho, b1, b2):
        """
        Initial method for Tri31

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of three element nodes in counter-clockwise order
        thick: float
            Element thickness
        otype: str
            String representing material behavior. the type parameter can be either ``'planestrain'`` or
            ``'planestress'``
        mat: obj
            Object of ndmaterial
        pressure: float
            Surface pressure (optional, default = 0.0)
        rho: float
            Element mass density (per unit volume) from which a lumped element mass matrix is computed (optional,
            default=0.0)
        b1: float
            Constant body forces defined in the domain (optional, default=0.0)
        b2: float
            Constant body forces defined in the domain (optional, default=0.0)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0], [1, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(3)]
        >>> mat = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
        >>> o3.element.Tri31(osi, ele_nodes=ele_nodes, thick=1.0, otype=o3.cc.PLANE_STRAIN, mat=mat, pressure=1.0, rho=1.0, b1=1.0, b2=1.0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.thick = float(thick)
        self.otype = otype
        self.mat = mat
        self.pressure = float(pressure)
        self.rho = float(rho)
        self.b1 = float(b1)
        self.b2 = float(b2)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.thick, self.otype, self.mat.tag, self.pressure, self.rho, self.b1, self.b2]
        self.to_process(osi)



import o3seespy as o3  # for testing only


def test_coulomb():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.friction_model.Coulomb(osi, mu=1.0)


def test_vel_dependent():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.friction_model.VelDependent(osi, mu_slow=1.0, mu_fast=1.0, trans_rate=1.0)


def test_vel_normal_frc_dep():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.friction_model.VelNormalFrcDep(osi, a_slow=1.0, n_slow=1.0, a_fast=1.0, n_fast=1.0, alpha0=1.0, alpha1=1.0, alpha2=1.0, max_mu_fact=1.0)


def test_vel_pressure_dep():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.friction_model.VelPressureDep(osi, mu_slow=1.0, mu_fast0=1.0, big_a=1.0, delta_mu=1.0, alpha=1.0, trans_rate=1.0)


def test_vel_dep_multi_linear():
    osi = o3.OpenSeesInstance(ndm=2)
    vel_points = [0.0, 1.0]
    frn_points = [1.0, 1.0]
    o3.friction_model.VelDepMultiLinear(osi, vel_points=vel_points, frn_points=frn_points)


if __name__ == '__main__':
    test_vel_dep_multi_linear()



import o3seespy as o3  # for testing only
import pytest

@pytest.mark.skip()
def test_fluid_solid_porous_material():
    osi = o3.OpenSeesInstance(ndm=2, ndf=2)
    soil_mat = o3.nd_material.ElasticIsotropic(osi, e_mod=1.0, nu=1.0, rho=0.0)
    o3.nd_material.FluidSolidPorousMaterial(osi, nd=1.0, soil_mat=soil_mat, combined_bulk_modul=1.0, pa=101.0)




import o3seespy as o3  # for testing only
import pytest


def test_surface_load():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
    o3.element.SurfaceLoad(osi, ele_nodes=ele_nodes, p=1.0)


def test_vs3d4():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
    o3.element.VS3D4(osi, ele_nodes=ele_nodes, big_e=1.0, big_g=1.0, rho=1.0, big_r=1.0, alpha_n=1.0, alpha_t=1.0)


@pytest.mark.skip()
def test_ac3d8():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
    mat = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
    o3.element.AC3D8(osi, ele_nodes=ele_nodes, mat=mat)


@pytest.mark.skip()
def test_asi3d8():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.element.ASI3D8(osi, ele_nodes1=1, ele_nodes2=1)


@pytest.mark.skip()
def test_av3d4():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
    mat = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
    o3.element.AV3D4(osi, ele_nodes=ele_nodes, mat=mat)




import o3seespy as o3  # for testing only
import pytest


def test_quad():
    osi = o3.OpenSeesInstance(ndm=2, ndf=2)
    obj = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
    coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
    o3.element.Quad(osi, ele_nodes=ele_nodes, thick=1.0, otype='PlaneStrain', mat=obj, pressure=1.0, rho=1.0, b1=0.0, b2=0.0)



def test_shell_mitc4():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    sec = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
    o3.element.ShellMITC4(osi, ele_nodes=ele_nodes, sec=sec)


def test_shell_dkgq():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    sec = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
    o3.element.ShellDKGQ(osi, ele_nodes=ele_nodes, sec=sec)


@pytest.mark.skip()  # this is been fixed by needs latest openseespy
def test_shell_dkgt():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0], [1, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    sec = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
    o3.element.ShellDKGT(osi, ele_nodes=ele_nodes, sec=sec)


def test_shell_nldkgq():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    sec = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
    o3.element.ShellNLDKGQ(osi, ele_nodes=ele_nodes, sec=sec)


@pytest.mark.skip()
def test_shell_nldkgt():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0], [1, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    sec = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
    o3.element.ShellNLDKGT(osi, ele_nodes=ele_nodes, sec=sec)


def test_shell_nl():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0], [1, 1], [0, 1], [0.5, 0], [1, 0.5], [0.5, 1], [0, 0.5], [0.5, 0.5]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    sec = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
    o3.element.ShellNL(osi, ele_nodes=ele_nodes, sec=sec)


def test_bbar_quad():
    osi = o3.OpenSeesInstance(ndm=2, ndf=2)
    mat = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
    coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
    o3.element.BbarQuad(osi, ele_nodes=ele_nodes, thick=1.0, mat=mat)


def test_enhanced_quad():
    osi = o3.OpenSeesInstance(ndm=2, ndf=2)
    obj = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
    coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
    o3.element.EnhancedQuad(osi, ele_nodes=ele_nodes, thick=1.0, otype='PlaneStress', mat=obj)


def test_ss_pquad():
    osi = o3.OpenSeesInstance(ndm=2)
    obj = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
    coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
    o3.element.SSPquad(osi, ele_nodes=ele_nodes, mat=obj, otype='PlaneStrain', thick=1.0, b1=0.0, b2=0.0)


if __name__ == '__main__':
    test_bbar_quad()




from setuptools import setup, find_packages

about = {}
with open("o3seespy/__about__.py") as fp:
    exec(fp.read(), about)

with open('README.rst') as readme_file:
    readme = readme_file.read()

# with open('HISTORY.rst') as history_file:
#     history = history_file.read()

setup(name='o3seespy',
      version=about['__version__'],
      description='Object-oriented native Python version of OpenSees',
      long_description=readme,  # + '\n\n' + history,
      url='',
      author=about['__author__'],
      author_email='mmi46@uclive.ac.nz',
      license=about['__license__'],
      package_data={
            'o3seespy': ['o3seespy/command/mat_recorder_options.csv'],
        },
      include_package_data=True,
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Intended Audience :: Science/Research',
          'Topic :: Scientific/Engineering',
          'Programming Language :: Python :: 3',
      ],
      packages=find_packages(exclude=['contrib', 'docs', 'tests', 'examples',
                                      '_auto_build', '.github']),
      install_requires=[
        "numpy",  # now it is optional
        'openseespy>=3.1.5.12 ; sys_platform!="darwin"',
        'openseespy==3.1.5.12 ; sys_platform=="darwin"',

      ],
      # List additional groups of dependencies here (e.g. development
      # dependencies). You can install these using the following syntax,
      # for example:
      # $ pip install -e .[dev,test]
      extras_require={
          'test': ['pytest'],
      },
      python_requires='>=3',
      zip_safe=False)


# From python packaging guides
# versioning is a 3-part MAJOR.MINOR.MAINTENANCE numbering scheme,
# where the project author increments:

# MAJOR version when they make incompatible API changes,
# MINOR version when they add functionality in a backwards-compatible manner, and
# MAINTENANCE version when they make backwards-compatible bug fixes.


from o3seespy.command.element.base_element import ElementBase


class SurfaceLoad(ElementBase):
    """
    The SurfaceLoad Element Class
    
    This command is used to construct a SurfaceLoad element object.

    
    """
    op_type = 'SurfaceLoad'

    def __init__(self, osi, ele_nodes, p):
        """
        Initial method for SurfaceLoad

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            The four nodes defining the element, input in counterclockwise order (-ndm 3 -ndf 3)
        p: float
            Applied pressure loading normal to the surface, outward is positive, inward is negative

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
        >>> o3.element.SurfaceLoad(osi, ele_nodes=ele_nodes, p=1.0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.p = float(p)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.p]
        self.to_process(osi)


class VS3D4(ElementBase):
    """
    The VS3D4 Element Class
    
    This command is used to construct a four-node 3D viscous-spring boundary quad element object based on a bilinear
    isoparametric formulation.

    
    """
    op_type = 'VS3D4'

    def __init__(self, osi, ele_nodes, big_e, big_g, rho, big_r, alpha_n, alpha_t):
        """
        Initial method for VS3D4

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            4 end nodes
        big_e: float
            Young's modulus of element material
        big_g: float
            Shear modulus of element material
        rho: float
            Mass density of element material
        big_r: float
            Distance from the scattered wave source to the boundary
        alpha_n: float
            Correction parameter in the normal direction
        alpha_t: float
            Correction parameter in the tangential direction

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
        >>> o3.element.VS3D4(osi, ele_nodes=ele_nodes, big_e=1.0, big_g=1.0, rho=1.0, big_r=1.0, alpha_n=1.0, alpha_t=1.0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.big_e = float(big_e)
        self.big_g = float(big_g)
        self.rho = float(rho)
        self.big_r = float(big_r)
        self.alpha_n = float(alpha_n)
        self.alpha_t = float(alpha_t)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.big_e, self.big_g, self.rho, self.big_r, self.alpha_n, self.alpha_t]
        self.to_process(osi)


class AC3D8(ElementBase):
    """
    The AC3D8 Element Class
    
    This command is used to construct an eight-node 3D brick acoustic element object based on a trilinear isoparametric
    formulation.

    
    """
    op_type = 'AC3D8'

    def __init__(self, osi, ele_nodes, mat):
        """
        Initial method for AC3D8

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            8 end nodes
        mat: obj
            Material object of previously defined nd material

        Examples
        --------
        >>> import o3seespy as o3
        >>> # Example is currently not working
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
        >>> mat = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
        >>> o3.element.AC3D8(osi, ele_nodes=ele_nodes, mat=mat)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.mat = mat
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.mat.tag]
        self.to_process(osi)


class ASI3D8(ElementBase):
    """
    The ASI3D8 Element Class
    
    This command is used to construct an eight-node zero-thickness 3D brick acoustic-structure interface element object
    based on a bilinear isoparametric formulation. The nodes in the acoustic domain share the same coordinates with the
    nodes in the solid domain.

    
    """
    op_type = 'ASI3D8'

    def __init__(self, osi, ele_nodes1, ele_nodes2):
        """
        Initial method for ASI3D8

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes1: None
            
        ele_nodes2: None
            

        Examples
        --------
        >>> import o3seespy as o3
        >>> # Example is currently not working
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.element.ASI3D8(osi, ele_nodes1=1, ele_nodes2=1)
        """
        self.osi = osi
        self.ele_nodes1 = ele_nodes1
        self.ele_nodes2 = ele_nodes2
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_nodes1, *self.ele_nodes2]
        self.to_process(osi)


class AV3D4(ElementBase):
    """
    The AV3D4 Element Class
    
    This command is used to construct a four-node 3D acoustic viscous boundary quad element object based on a bilinear
    isoparametric formulation.

    
    """
    op_type = 'AV3D4'

    def __init__(self, osi, ele_nodes, mat):
        """
        Initial method for AV3D4

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            4 end nodes
        mat: obj
            Material object of previously defined nd material

        Examples
        --------
        >>> import o3seespy as o3
        >>> # Example is currently not working
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
        >>> mat = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
        >>> o3.element.AV3D4(osi, ele_nodes=ele_nodes, mat=mat)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.mat = mat
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.mat.tag]
        self.to_process(osi)



FREE = 0
FIXED = 1

X = 1  # deprecated
Y = 2  # deprecated
ROTZ = 3  # deprecated
DOF_X = 1
DOF_Y = 2
DOF2D_X = 1
DOF2D_Y = 2
DOF2D_ROTZ = 3
DOF2D_PP = 3  # pore pressure
DOF3D_Z = 3
DOF3D_ROTX = 4
DOF3D_ROTY = 5
DOF3D_ROTZ = 6

P = 'P'
M_Z = 'Mz'
V_Y = 'Vy'
M_Y = 'My'
V_Z = 'Vz'
TORSION = 'T'

PLANE_STRAIN = "PlaneStrain"
PLANE_STRESS = "PlaneStress"

# DOFD = {1: ['X'],
#         2: ['X', 'Y'],
#         3: ['X', 'ROTZ']}



from o3seespy.base_model import OpenSeesObject


class MeshBase(OpenSeesObject):
    op_base_type = "mesh"


class Line(MeshBase):
    """
    The Line Mesh Class
    
    Create a line mesh object. 
    """
    op_type = 'line'

    def __init__(self, osi, numnodes, ndtags, id, ndf, meshsize, ele_type='', ele_args: float=None):
        """
        Initial method for Line

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        numnodes: int
            Number of nodes for defining consective lines.
        ndtags: list
            The node objects
        id: int
            Mesh id. meshes with same id are considered as same structure of fluid identity. * ``id`` = 0 : not in fsi *
            ``id`` > 0 : structure * ``id`` < 0 : fluid
        ndf: int
            Ndf for nodes to be created.
        meshsize: float
            Mesh size.
        ele_type: str, optional
            The type of the element,  * :ref:`elasticbeamcolumn` * :ref:`forcebeamcolumn-element` *
            :ref:`dispbeamcolumn-element` if no type is given, only nodes are created
        ele_args: list (default=True), optional
            A list of element arguments. the arguments are same as in the element commands, but without element object,
            and node objects.  for example, ``eleargs = ['elasticbeamcolumn', a, e, iz, transfobject]``
        """
        self.osi = osi
        self.numnodes = int(numnodes)
        self.ndtags = ndtags
        self.id = int(id)
        self.ndf = int(ndf)
        self.meshsize = float(meshsize)
        self.ele_type = ele_type
        self.ele_args = ele_args
        osi.n_mesh += 1
        self._tag = osi.n_mesh
        self._parameters = [self.op_type, self._tag, self.numnodes, *self.ndtags, self.id, self.ndf, self.meshsize, self.ele_type]
        special_pms = ['ele_args']
        packets = [True]
        for i, pm in enumerate(special_pms):
            if getattr(self, pm) is not None:
                if packets[i]:
                    self._parameters += [*getattr(self, pm)]
                else:
                    self._parameters += [getattr(self, pm)]
            else:
                break
        self.to_process(osi)


class Tri(MeshBase):
    """
    The Tri Mesh Class
    
    Create a triangular mesh object.
    """
    op_type = 'tri'

    def __init__(self, osi, numlines, ltags, id, ndf, meshsize, ele_type='', ele_args: float=None):
        """
        Initial method for Tri

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        numlines: int
            Number of lines (:ref:`linemesh`) for defining a polygon.
        ltags: list
            The :ref:`linemesh` objects
        id: int
            Mesh id. meshes with same id are considered as same structure of fluid identity. * ``id`` = 0 : not in fsi *
            ``id`` > 0 : structure * ``id`` < 0 : fluid
        ndf: int
            Ndf for nodes to be created.
        meshsize: float
            Mesh size.
        ele_type: str, optional
            The element type,  * :doc:`pfemelementbubble` * :doc:`pfemelementcompressible` * :doc:`tri31` *
            :doc:`elasticbeamcolumn` * :doc:`forcebeamcolumn` * :doc:`dispbeamcolumn` if no type is given, only nodes
            are created. if beam elements are given, beams are created instead of triangular elements.
        ele_args: list (default=True), optional
            A list of element arguments. the arguments are same as in the element commands, but without element object,
            and node objects.  for example, ``eleargs = ['pfemelementbubble', rho, mu, b1, b2, thickness, kappa]``
        """
        self.osi = osi
        self.numlines = int(numlines)
        self.ltags = ltags
        self.id = int(id)
        self.ndf = int(ndf)
        self.meshsize = float(meshsize)
        self.ele_type = ele_type
        self.ele_args = ele_args
        osi.n_mesh += 1
        self._tag = osi.n_mesh
        self._parameters = [self.op_type, self._tag, self.numlines, *self.ltags, self.id, self.ndf, self.meshsize, self.ele_type]
        special_pms = ['ele_args']
        packets = [True]
        for i, pm in enumerate(special_pms):
            if getattr(self, pm) is not None:
                if packets[i]:
                    self._parameters += [*getattr(self, pm)]
                else:
                    self._parameters += [getattr(self, pm)]
            else:
                break
        self.to_process(osi)


class Quad(MeshBase):
    """
    The Quad Mesh Class
    
    Create a quad mesh object. The number of lines must be 4. These lines are continuousto form a loop.
    """
    op_type = 'quad'

    def __init__(self, osi, numlines, ltags, id, ndf, meshsize, ele_type='', ele_args: float=None):
        """
        Initial method for Quad

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        numlines: int
            Number of lines (:ref:`linemesh`) for defining a polygon.
        ltags: list
            The :ref:`linemesh` objects
        id: int
            Mesh id. meshes with same id are considered as same structure of fluid identity. * ``id`` = 0 : not in fsi *
            ``id`` > 0 : structure * ``id`` < 0 : fluid
        ndf: int
            Ndf for nodes to be created.
        meshsize: float
            Mesh size.
        ele_type: str, optional
            The element type,  * :doc:`pfemelementbubble` * :doc:`pfemelementcompressible` * :doc:`tri31` *
            :doc:`elasticbeamcolumn` * :doc:`forcebeamcolumn` * :doc:`dispbeamcolumn` * :doc:`shellmitc4` if no type
            is given, only nodes are created. if beam elements are given, beams are created instead of quad
            elements. if triangular elements are given, they are created by dividing one quad to two triangles.
        ele_args: list (default=True), optional
            A list of element arguments. the arguments are same as in the element commands, but without element object,
            and node objects.  for example, ``eleargs = ['pfemelementbubble', rho, mu, b1, b2, thickness, kappa]``
        """
        self.osi = osi
        self.numlines = int(numlines)
        self.ltags = ltags
        self.id = int(id)
        self.ndf = int(ndf)
        self.meshsize = float(meshsize)
        self.ele_type = ele_type
        self.ele_args = ele_args
        osi.n_mesh += 1
        self._tag = osi.n_mesh
        self._parameters = [self.op_type, self._tag, self.numlines, *self.ltags, self.id, self.ndf, self.meshsize, self.ele_type]
        special_pms = ['ele_args']
        packets = [True]
        for i, pm in enumerate(special_pms):
            if getattr(self, pm) is not None:
                if packets[i]:
                    self._parameters += [*getattr(self, pm)]
                else:
                    self._parameters += [getattr(self, pm)]
            else:
                break
        self.to_process(osi)


class Tet(MeshBase):
    """
    The Tet Mesh Class
    
    Create a 3D tetrahedron mesh object.
    """
    op_type = 'tet'

    def __init__(self, osi, nummesh, mtags, id, ndf, meshsize, ele_type='', ele_args: float=None):
        """
        Initial method for Tet

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        nummesh: int
            Number of 2d mesh for defining a 3d body.
        mtags: list
            The mesh objects
        id: int
            Mesh id. meshes with same id are considered as same structure of fluid identity. * ``id`` = 0 : not in fsi *
            ``id`` > 0 : structure * ``id`` < 0 : fluid
        ndf: int
            Ndf for nodes to be created.
        meshsize: float
            Mesh size.
        ele_type: str, optional
            The element type,  * :doc:`fournodetetrahedron` if no type is given, only nodes are created.
        ele_args: list (default=True), optional
            A list of element arguments. the arguments are same as in the element commands, but without element object,
            and node objects. 
        """
        self.osi = osi
        self.nummesh = int(nummesh)
        self.mtags = mtags
        self.id = int(id)
        self.ndf = int(ndf)
        self.meshsize = float(meshsize)
        self.ele_type = ele_type
        self.ele_args = ele_args
        osi.n_mesh += 1
        self._tag = osi.n_mesh
        self._parameters = [self.op_type, self._tag, self.nummesh, *self.mtags, self.id, self.ndf, self.meshsize, self.ele_type]
        special_pms = ['ele_args']
        packets = [True]
        for i, pm in enumerate(special_pms):
            if getattr(self, pm) is not None:
                if packets[i]:
                    self._parameters += [*getattr(self, pm)]
                else:
                    self._parameters += [getattr(self, pm)]
            else:
                break
        self.to_process(osi)


class Partvel(MeshBase):
    """
    The Partvel Mesh Class
    
    Create or return a group of particles which will be used for background mesh.
    """
    op_type = 'part'

    def __init__(self, osi, otype, p_args, vel0, ele_type='', ele_args: float=None):
        """
        Initial method for Partvel

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        otype: str
            Type of the mesh
        p_args: list
            Coordinates of points defining the mesh region and z directions * ``'quad'`` : [x1, y1, x2, y2, x3, y3, x4,
            y4, nx, ny] coordinates of four corners in counter-clock wise order. * ``'cube'`` : [x1, y1, z1, x2, y2, z2, x3, y3,
            z3, x4, y4, z4, x5, y5, z5, x6, y6, z6, x7, y7, z7, x8, y8, z8, nx, ny, nz] coordinates of four corners at bottom
            and at top in counter-clock wise order * ``'tri'`` : [x1, y1, x2, y2, x3, y3, nx, ny] coordinates of three
            corners in counter-clock wise order * ``'line'`` : [x1, y1, x2, y2, nx] coordinates of two ends in
            counter-clock wise order * ``'pointlist'`` : [x1, y1, <z1>, vx1, vy1, <vz1>, ax1, ay1, <az1>, p1,
             x2, y2, <z2>, vx2, vy2, <vz2>, ax2, ay2, <az2>, p2, ..] input particles' data in a list, in the
            order of coordinates of last time step, current coordinates, velocity, acceleration, and
            pressure. * ``'pointlist'`` without list return a list of current particles' data in
            this mesh [object1, x1, y1, <z1>, vx1, vy1, <vz1>, ax1, ay1, <az1>, p1, object2,
            x2, y2, <z2>, vx2, vy2, <vz2>, ax2, ay2, <az2>, p2, ..] the format is similar
            to the input list, but with an additional object for each particle.
        vel0: list
            A list of initial velocities. 
        ele_type: str, optional
            The element type,  * :doc:`pfemelementbubble` * :doc:`pfemelementcompressible` * :doc:`tri31` if no type is
            given, only nodes are created
        ele_args: list (default=True), optional
            A list of element arguments. (optional, see :doc:`linemesh` and :doc:`trimesh`)
        """
        self.osi = osi
        self.otype = otype
        self.p_args = p_args
        self.ele_type = ele_type
        self.ele_args = ele_args
        self.vel0 = vel0
        osi.n_mesh += 1
        self._tag = osi.n_mesh
        self._parameters = [self.op_type, self._tag, self.otype, *self.p_args, self.ele_type]
        special_pms = ['ele_args', 'vel0', 'p0']
        packets = [True, True, False]
        for i, pm in enumerate(special_pms):
            if getattr(self, pm) is not None:
                if packets[i]:
                    self._parameters += [*getattr(self, pm)]
                else:
                    self._parameters += [getattr(self, pm)]
            else:
                break
        self.to_process(osi)

class Partpressure(MeshBase):
    """
    The Partpressure Mesh Class
    
    Create or return a group of particles which will be used for background mesh.
    """
    op_type = 'part'

    def __init__(self, osi, otype, p_args, p0, ele_type='', ele_args: float=None):
        """
        Initial method for Partpressure

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        otype: str
            Type of the mesh
        p_args: list
            Coordinates of points defining the mesh region and z directions * ``'quad'`` : [x1, y1, x2, y2, x3, y3, x4,
            y4, nx, ny] coordinates of four corners in counter-clock wise order. * ``'cube'`` : [x1, y1, z1, x2, y2, z2, x3, y3,
            z3, x4, y4, z4, x5, y5, z5, x6, y6, z6, x7, y7, z7, x8, y8, z8, nx, ny, nz] coordinates of four corners at bottom
            and at top in counter-clock wise order * ``'tri'`` : [x1, y1, x2, y2, x3, y3, nx, ny] coordinates of three
            corners in counter-clock wise order * ``'line'`` : [x1, y1, x2, y2, nx] coordinates of two ends in
            counter-clock wise order * ``'pointlist'`` : [x1, y1, <z1>, vx1, vy1, <vz1>, ax1, ay1, <az1>, p1,
             x2, y2, <z2>, vx2, vy2, <vz2>, ax2, ay2, <az2>, p2, ..] input particles' data in a list, in the
            order of coordinates of last time step, current coordinates, velocity, acceleration, and
            pressure. * ``'pointlist'`` without list return a list of current particles' data in
            this mesh [object1, x1, y1, <z1>, vx1, vy1, <vz1>, ax1, ay1, <az1>, p1, object2,
            x2, y2, <z2>, vx2, vy2, <vz2>, ax2, ay2, <az2>, p2, ..] the format is similar
            to the input list, but with an additional object for each particle.
        p0: float
            Initial pressure. 
        ele_type: str, optional
            The element type,  * :doc:`pfemelementbubble` * :doc:`pfemelementcompressible` * :doc:`tri31` if no type is
            given, only nodes are created
        ele_args: list (default=True), optional
            A list of element arguments. (optional, see :doc:`linemesh` and :doc:`trimesh`)
        """
        self.osi = osi
        self.otype = otype
        self.p_args = p_args
        self.ele_type = ele_type
        self.ele_args = ele_args
        self.p0 = float(p0)
        osi.n_mesh += 1
        self._tag = osi.n_mesh
        self._parameters = [self.op_type, self._tag, self.otype, *self.p_args, self.ele_type]
        special_pms = ['ele_args', 'vel0', 'p0']
        packets = [True, True, False]
        for i, pm in enumerate(special_pms):
            if getattr(self, pm) is not None:
                if packets[i]:
                    self._parameters += [*getattr(self, pm)]
                else:
                    self._parameters += [getattr(self, pm)]
            else:
                break
        self.to_process(osi)


class Bgwave(MeshBase):
    """
    The Bgwave Mesh Class
    
    Create a background mesh. 
    """
    op_type = 'bg'

    def __init__(self, osi, lower, upper, wavefilename, numl, locations, tol: float=None, meshtol: float=None, numsub=None):
        """
        Initial method for Bgwave

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        lower: list
            A list of coordinates of the lower point of the background region.
        upper: list
            A list of coordinates of the uuper point of the background region.
        wavefilename: str
            A filename to record wave heights and velocities 
        numl: int
            Number of locations to record wave 
        locations: list
            Coordinates of the locations 
        tol: float, optional
            Tolerance for intri check. (optional, default 1e-10)
        meshtol: float, optional
            Tolerance for cell boundary check. (optional, default 0.1)
        numsub: None, optional
            
        """
        self.osi = osi
        self.lower = lower
        self.upper = upper
        if tol is None:
            self.tol = None
        else:
            self.tol = float(tol)
        if meshtol is None:
            self.meshtol = None
        else:
            self.meshtol = float(meshtol)
        self.wavefilename = wavefilename
        self.numl = int(numl)
        self.locations = locations
        self.numsub = numsub
        osi.n_mesh += 1
        self._tag = osi.n_mesh
        self._parameters = [self.op_type, self._tag, *self.lower, *self.upper, '-wave', self.wavefilename, self.numl, *self.locations]
        if getattr(self, 'tol') is not None:
            self._parameters += ['-tol', self.tol]
        if getattr(self, 'meshtol') is not None:
            self._parameters += ['-meshtol', self.meshtol]
        if getattr(self, 'numsub') is not None:
            self._parameters += ['-numsub', self.numsub]
        self.to_process(osi)

class Bgstructure(MeshBase):
    """
    The Bgstructure Mesh Class
    
    Create a background mesh. 
    """
    op_type = 'bg'

    def __init__(self, osi, lower, upper, id, numnodes, snodes, tol: float=None, meshtol: float=None):
        """
        Initial method for Bgstructure

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        lower: list
            A list of coordinates of the lower point of the background region.
        upper: list
            A list of coordinates of the uuper point of the background region.
        id: int
            Structural id > 0, same meaning as :doc:`trimesh` 
        numnodes: None
            
        snodes: None
            
        tol: float, optional
            Tolerance for intri check. (optional, default 1e-10)
        meshtol: float, optional
            Tolerance for cell boundary check. (optional, default 0.1)
        """
        self.osi = osi
        self.lower = lower
        self.upper = upper
        if tol is None:
            self.tol = None
        else:
            self.tol = float(tol)
        if meshtol is None:
            self.meshtol = None
        else:
            self.meshtol = float(meshtol)
        self.id = int(id)
        self.numnodes = numnodes
        self.snodes = snodes
        osi.n_mesh += 1
        self._tag = osi.n_mesh
        self._parameters = [self.op_type, self._tag, *self.lower, *self.upper, '-structure', self.id, self.numnodes, *self.snodes]
        if getattr(self, 'tol') is not None:
            self._parameters += ['-tol', self.tol]
        if getattr(self, 'meshtol') is not None:
            self._parameters += ['-meshtol', self.meshtol]
        self.to_process(osi)

class BglargeSize(MeshBase):
    """
    The BglargeSize Mesh Class
    
    Create a background mesh. 
    """
    op_type = 'bg'

    def __init__(self, osi, lower, upper, level, llower, lupper, tol: float=None, meshtol: float=None):
        """
        Initial method for BglargeSize

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        lower: list
            A list of coordinates of the lower point of the background region.
        upper: list
            A list of coordinates of the uuper point of the background region.
        level: int
            Some regions can have larger mesh size with larger ``level``. ``level = 1`` means same as basic mesh size.
        llower: list
            A list of coordinates of the lower point of the region with larger mesh size 
        lupper: list
            A list of coordinates of the upper point of the region with larger mesh size
        tol: float, optional
            Tolerance for intri check. (optional, default 1e-10)
        meshtol: float, optional
            Tolerance for cell boundary check. (optional, default 0.1)
        """
        self.osi = osi
        self.lower = lower
        self.upper = upper
        if tol is None:
            self.tol = None
        else:
            self.tol = float(tol)
        if meshtol is None:
            self.meshtol = None
        else:
            self.meshtol = float(meshtol)
        self.level = int(level)
        self.llower = llower
        self.lupper = lupper
        osi.n_mesh += 1
        self._tag = osi.n_mesh
        self._parameters = [self.op_type, self._tag, *self.lower, *self.upper, '-largeSize', self.level, *self.llower, *self.lupper]
        if getattr(self, 'tol') is not None:
            self._parameters += ['-tol', self.tol]
        if getattr(self, 'meshtol') is not None:
            self._parameters += ['-meshtol', self.meshtol]
        self.to_process(osi)



import o3seespy as o3
import pytest


def test_can_set_PressureIndependMultiYield():
    osi = o3.OpenSeesInstance(ndm=2)

    # Define material
    nd = 2
    rho = 1400
    ref_shear_modul = 40e6
    ref_bulk_modul = 60e6
    cohesi = 60e3
    peak_shear_stra = 0.02
    o3.nd_material.PressureIndependMultiYield(osi, nd, rho, ref_shear_modul, ref_bulk_modul,
                                               cohesi, peak_shear_stra)


def test_can_set_PM4Sand():
    osi = o3.OpenSeesInstance(ndm=2)

    # Define material
    d_r = 0.5
    hpo = 0.4
    g0 = 555.
    p_atm = 101.3
    rho = 1.7
    o3.nd_material.PM4Sand(osi, d_r, g0, hpo, rho, p_atm)


def skip_test_can_set_stress_density_model():
    osi = o3.OpenSeesInstance(ndm=2)

    # Define material
    o3.nd_material.StressDensityModel(osi, den=1.8, e_init=0.73, area=250., n=0.6, nu=0.3, a1=0.58, b1=0.023,
                                       a2=230.0, b2=65., a3=79., b3=16., fd=4.0, mu_not=0.22, mu_cyc=0.0, sc=0.0055,
                                       big_m=0.607, p_atm=98.1)



import o3seespy as o3  # for testing only


def test_contact_material2d():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.nd_material.ContactMaterial2D(osi, mu=1.0, g_mod=1.0, c=1.0, t=1.0)


def test_contact_material3d():
    osi = o3.OpenSeesInstance(ndm=3)
    o3.nd_material.ContactMaterial3D(osi, mu=1.0, g_mod=1.0, c=1.0, t=1.0)




from o3seespy.base_model import OpenSeesObject


class TestBase(OpenSeesObject):
    op_base_type = "test"


class NormUnbalance(TestBase):
    """
    The NormUnbalance Test Class
    
    Create a NormUnbalance test, which uses the norm of the right hand side of the matrix equation to determine if
    convergence has been reached.
    """
    op_type = 'NormUnbalance'

    def __init__(self, osi, tol, max_iter, p_flag=0, n_type=2, max_incr: float=None):
        r"""
        Initial method for NormUnbalance

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        tol: float
            Tolerance criteria used to check for convergence.
        max_iter: int
            Max number of iterations to check
        p_flag: int, optional
            Print flag : * 0 print nothing. * 1 print information on norms each time ``test()`` is invoked. * 2 print
            information on norms and number of iterations at end of successful test. * 4 at each step it will print the norms
            and also the :math:`\delta u` and :math:`r(u)` vectors. * 5 if it fails to converge at end of ``numiter`` it
            will print an error message **but return a successfull test**.
        n_type: int, optional
            Type of norm, (0 = max-norm, 1 = 1-norm, 2 = 2-norm). 
        max_incr: int (default=True), optional
            Maximum times of error increasing. 

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.test.NormUnbalance(osi, tol=1.0, max_iter=1, p_flag=0, n_type=2, max_incr=None)
        """
        self.osi = osi
        self.tol = float(tol)
        self.max_iter = int(max_iter)
        self.p_flag = int(p_flag)
        self.n_type = int(n_type)
        if max_incr is None:
            self.max_incr = None
        else:
            self.max_incr = int(max_incr)
        self._parameters = [self.op_type, self.tol, self.max_iter, self.p_flag, self.n_type]
        special_pms = ['max_incr']
        packets = [False]
        for i, pm in enumerate(special_pms):
            if getattr(self, pm) is not None:
                if packets[i]:
                    self._parameters += [*getattr(self, pm)]
                else:
                    self._parameters += [getattr(self, pm)]
            else:
                break
        self.to_process(osi)


class NormDispIncr(TestBase):
    """
    The NormDispIncr Test Class
    
    Create a NormUnbalance test, which uses the norm of the left hand side solution vector of the matrix equation to
    determine if convergence has been reached.
    """
    op_type = 'NormDispIncr'

    def __init__(self, osi, tol, max_iter, p_flag=0, n_type=2):
        r"""
        Initial method for NormDispIncr

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        tol: float
            Tolerance criteria used to check for convergence.
        max_iter: int
            Max number of iterations to check
        p_flag: int, optional
            Print flag : * 0 print nothing. * 1 print information on norms each time ``test()`` is invoked. * 2 print
            information on norms and number of iterations at end of successful test. * 4 at each step it will print the norms
            and also the :math:`\delta u` and :math:`r(u)` vectors. * 5 if it fails to converge at end of ``numiter`` it
            will print an error message **but return a successfull test**.
        n_type: int, optional
            Type of norm, (0 = max-norm, 1 = 1-norm, 2 = 2-norm). 

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.test.NormDispIncr(osi, tol=1.0, max_iter=1, p_flag=0, n_type=2)
        """
        self.osi = osi
        self.tol = float(tol)
        self.max_iter = int(max_iter)
        self.p_flag = int(p_flag)
        self.n_type = int(n_type)
        self._parameters = [self.op_type, self.tol, self.max_iter, self.p_flag, self.n_type]
        self.to_process(osi)


class EnergyIncr(TestBase):
    """
    The EnergyIncr Test Class
    
    Create a EnergyIncr test, which uses the dot product of the solution vector and norm of the right hand side of the
    matrix equation to determine if convergence has been reached.
    """
    op_type = 'EnergyIncr'

    def __init__(self, osi, tol, max_iter, p_flag=0, n_type=2):
        r"""
        Initial method for EnergyIncr

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        tol: float
            Tolerance criteria used to check for convergence.
        max_iter: int
            Max number of iterations to check
        p_flag: int, optional
            Print flag : * 0 print nothing. * 1 print information on norms each time ``test()`` is invoked. * 2 print
            information on norms and number of iterations at end of successful test. * 4 at each step it will print the norms
            and also the :math:`\delta u` and :math:`r(u)` vectors. * 5 if it fails to converge at end of ``numiter`` it
            will print an error message **but return a successfull test**.
        n_type: int, optional
            Type of norm, (0 = max-norm, 1 = 1-norm, 2 = 2-norm). 

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.test.EnergyIncr(osi, tol=1.0, max_iter=1, p_flag=0, n_type=2)
        """
        self.osi = osi
        self.tol = float(tol)
        self.max_iter = int(max_iter)
        self.p_flag = int(p_flag)
        self.n_type = int(n_type)
        self._parameters = [self.op_type, self.tol, self.max_iter, self.p_flag, self.n_type]
        self.to_process(osi)


class RelativeNormUnbalance(TestBase):
    """
    The RelativeNormUnbalance Test Class
    
    Create a RelativeNormUnbalance test, which uses the relative norm of the right hand side of the matrix equation to
    determine if convergence has been reached.
    """
    op_type = 'RelativeNormUnbalance'

    def __init__(self, osi, tol, max_iter, p_flag=0, n_type=2):
        r"""
        Initial method for RelativeNormUnbalance

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        tol: float
            Tolerance criteria used to check for convergence.
        max_iter: int
            Max number of iterations to check
        p_flag: int, optional
            Print flag : * 0 print nothing. * 1 print information on norms each time ``test()`` is invoked. * 2 print
            information on norms and number of iterations at end of successful test. * 4 at each step it will print the norms
            and also the :math:`\delta u` and :math:`r(u)` vectors. * 5 if it fails to converge at end of ``numiter`` it
            will print an error message **but return a successfull test**.
        n_type: int, optional
            Type of norm, (0 = max-norm, 1 = 1-norm, 2 = 2-norm). 

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.test.RelativeNormUnbalance(osi, tol=1.0, max_iter=1, p_flag=0, n_type=2)
        """
        self.osi = osi
        self.tol = float(tol)
        self.max_iter = int(max_iter)
        self.p_flag = int(p_flag)
        self.n_type = int(n_type)
        self._parameters = [self.op_type, self.tol, self.max_iter, self.p_flag, self.n_type]
        self.to_process(osi)


class RelativeNormDispIncr(TestBase):
    """
    The RelativeNormDispIncr Test Class
    
    Create a RelativeNormDispIncr test, which uses the relative of the solution vector of the matrix equation to
    determine if convergence has been reached.
    """
    op_type = 'RelativeNormDispIncr'

    def __init__(self, osi, tol, max_iter, p_flag=0, n_type=2):
        r"""
        Initial method for RelativeNormDispIncr

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        tol: float
            Tolerance criteria used to check for convergence.
        max_iter: int
            Max number of iterations to check
        p_flag: int, optional
            Print flag : * 0 print nothing. * 1 print information on norms each time ``test()`` is invoked. * 2 print
            information on norms and number of iterations at end of successful test. * 4 at each step it will print the norms
            and also the :math:`\delta u` and :math:`r(u)` vectors. * 5 if it fails to converge at end of ``numiter`` it
            will print an error message **but return a successfull test**.
        n_type: int, optional
            Type of norm, (0 = max-norm, 1 = 1-norm, 2 = 2-norm). 

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.test.RelativeNormDispIncr(osi, tol=1.0, max_iter=1, p_flag=0, n_type=2)
        """
        self.osi = osi
        self.tol = float(tol)
        self.max_iter = int(max_iter)
        self.p_flag = int(p_flag)
        self.n_type = int(n_type)
        self._parameters = [self.op_type, self.tol, self.max_iter, self.p_flag, self.n_type]
        self.to_process(osi)


class RelativeTotalNormDispIncr(TestBase):
    """
    The RelativeTotalNormDispIncr Test Class
    
    Create a RelativeTotalNormDispIncr test, which uses the ratio of the current norm to the total norm (the sum of all
    the norms since last convergence) of the solution vector.
    """
    op_type = 'RelativeTotalNormDispIncr'

    def __init__(self, osi, tol, max_iter, p_flag=0, n_type=2):
        r"""
        Initial method for RelativeTotalNormDispIncr

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        tol: float
            Tolerance criteria used to check for convergence.
        max_iter: int
            Max number of iterations to check
        p_flag: int, optional
            Print flag : * 0 print nothing. * 1 print information on norms each time ``test()`` is invoked. * 2 print
            information on norms and number of iterations at end of successful test. * 4 at each step it will print the norms
            and also the :math:`\delta u` and :math:`r(u)` vectors. * 5 if it fails to converge at end of ``numiter`` it
            will print an error message **but return a successfull test**.
        n_type: int, optional
            Type of norm, (0 = max-norm, 1 = 1-norm, 2 = 2-norm). 

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.test.RelativeTotalNormDispIncr(osi, tol=1.0, max_iter=1, p_flag=0, n_type=2)
        """
        self.osi = osi
        self.tol = float(tol)
        self.max_iter = int(max_iter)
        self.p_flag = int(p_flag)
        self.n_type = int(n_type)
        self._parameters = [self.op_type, self.tol, self.max_iter, self.p_flag, self.n_type]
        self.to_process(osi)


class RelativeEnergyIncr(TestBase):
    """
    The RelativeEnergyIncr Test Class
    
    Create a RelativeEnergyIncr test, which uses the relative dot product of the solution vector and norm of the right
    hand side of the matrix equation to determine if convergence has been reached.
    """
    op_type = 'RelativeEnergyIncr'

    def __init__(self, osi, tol, max_iter, p_flag=0, n_type=2):
        r"""
        Initial method for RelativeEnergyIncr

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        tol: float
            Tolerance criteria used to check for convergence.
        max_iter: int
            Max number of iterations to check
        p_flag: int, optional
            Print flag : * 0 print nothing. * 1 print information on norms each time ``test()`` is invoked. * 2 print
            information on norms and number of iterations at end of successful test. * 4 at each step it will print the norms
            and also the :math:`\delta u` and :math:`r(u)` vectors. * 5 if it fails to converge at end of ``numiter`` it
            will print an error message **but return a successfull test**.
        n_type: int, optional
            Type of norm, (0 = max-norm, 1 = 1-norm, 2 = 2-norm). 

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.test.RelativeEnergyIncr(osi, tol=1.0, max_iter=1, p_flag=0, n_type=2)
        """
        self.osi = osi
        self.tol = float(tol)
        self.max_iter = int(max_iter)
        self.p_flag = int(p_flag)
        self.n_type = int(n_type)
        self._parameters = [self.op_type, self.tol, self.max_iter, self.p_flag, self.n_type]
        self.to_process(osi)


class FixedNumIter(TestBase):
    """
    The FixedNumIter Test Class
    
    Create a FixedNumIter test, that performs a fixed number of iterations without testing for convergence.
    """
    op_type = 'FixedNumIter'

    def __init__(self, osi, max_iter, p_flag=0, n_type=2):
        r"""
        Initial method for FixedNumIter

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        max_iter: int
            Max number of iterations to check
        p_flag: int, optional
            Print flag : * 0 print nothing. * 1 print information on norms each time ``test()`` is invoked. * 2 print
            information on norms and number of iterations at end of successful test. * 4 at each step it will print the norms
            and also the :math:`\delta u` and :math:`r(u)` vectors. * 5 if it fails to converge at end of ``numiter`` it
            will print an error message **but return a successfull test**.
        n_type: int, optional
            Type of norm, (0 = max-norm, 1 = 1-norm, 2 = 2-norm). 

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.test.FixedNumIter(osi, max_iter=1, p_flag=0, n_type=2)
        """
        self.osi = osi
        self.max_iter = int(max_iter)
        self.p_flag = int(p_flag)
        self.n_type = int(n_type)
        self._parameters = [self.op_type, self.max_iter, self.p_flag, self.n_type]
        self.to_process(osi)


class NormDispAndUnbalance(TestBase):
    """
    The NormDispAndUnbalance Test Class
    
    Create a NormDispAndUnbalance test, which check if both
    """
    op_type = 'NormDispAndUnbalance'

    def __init__(self, osi, tol_incr, tol_r, max_iter, p_flag=0, n_type=2, maxincr=-1):
        r"""
        Initial method for NormDispAndUnbalance

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        tol_incr: float
            Tolerance for right hand residual
        tol_r: None
            
        max_iter: int
            Max number of iterations to check
        p_flag: int, optional
            Print flag : * 0 print nothing. * 1 print information on norms each time ``test()`` is invoked. * 2 print
            information on norms and number of iterations at end of successful test. * 4 at each step it will print the norms
            and also the :math:`\delta u` and :math:`r(u)` vectors. * 5 if it fails to converge at end of ``numiter`` it
            will print an error message **but return a successfull test**.
        n_type: int, optional
            Type of norm, (0 = max-norm, 1 = 1-norm, 2 = 2-norm). 
        maxincr: int, optional
            Maximum times of error increasing. 

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.test.NormDispAndUnbalance(osi, tol_incr=1.0, tol_r=1, max_iter=1, p_flag=0, n_type=2, maxincr=-1)
        """
        self.osi = osi
        self.tol_incr = float(tol_incr)
        self.tol_r = tol_r
        self.max_iter = int(max_iter)
        self.p_flag = int(p_flag)
        self.n_type = int(n_type)
        self.maxincr = int(maxincr)
        self._parameters = [self.op_type, self.tol_incr, self.tol_r, self.max_iter, self.p_flag, self.n_type, self.maxincr]
        self.to_process(osi)


class NormDispOrUnbalance(TestBase):
    """
    The NormDispOrUnbalance Test Class
    
    Create a NormDispOrUnbalance test, which check if both
    """
    op_type = 'NormDispOrUnbalance'

    def __init__(self, osi, tol_incr, tol_r, max_iter, p_flag=0, n_type=2, maxincr=-1):
        r"""
        Initial method for NormDispOrUnbalance

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        tol_incr: float
            Tolerance for right hand residual
        tol_r: None
            
        max_iter: int
            Max number of iterations to check
        p_flag: int, optional
            Print flag : * 0 print nothing. * 1 print information on norms each time ``test()`` is invoked. * 2 print
            information on norms and number of iterations at end of successful test. * 4 at each step it will print the norms
            and also the :math:`\delta u` and :math:`r(u)` vectors. * 5 if it fails to converge at end of ``numiter`` it
            will print an error message **but return a successfull test**.
        n_type: int, optional
            Type of norm, (0 = max-norm, 1 = 1-norm, 2 = 2-norm). 
        maxincr: int, optional
            Maximum times of error increasing. 

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.test.NormDispOrUnbalance(osi, tol_incr=1.0, tol_r=1, max_iter=1, p_flag=0, n_type=2, maxincr=-1)
        """
        self.osi = osi
        self.tol_incr = float(tol_incr)
        self.tol_r = tol_r
        self.max_iter = int(max_iter)
        self.p_flag = int(p_flag)
        self.n_type = int(n_type)
        self.maxincr = int(maxincr)
        self._parameters = [self.op_type, self.tol_incr, self.tol_r, self.max_iter, self.p_flag, self.n_type, self.maxincr]
        self.to_process(osi)



from o3seespy.base_model import OpenSeesObject


class AnalysisBase(OpenSeesObject):
    op_base_type = "analysis"


class Static(AnalysisBase):
    op_type = "Static"

    def __init__(self, osi):
        self.osi = osi
        self._parameters = [self.op_type]
        self.to_process(osi)


class Transient(AnalysisBase):
    op_type = "Transient"

    def __init__(self, osi):
        self.osi = osi
        self._parameters = [self.op_type]
        self.to_process(osi)


class VariableTransient(AnalysisBase):
    op_type = "VariableTransient"

    def __init__(self, osi):
        self.osi = osi
        self._parameters = [self.op_type]
        self.to_process(osi)


class PFEM(AnalysisBase):
    op_type = "PFEM"

    def __init__(self, osi, dt_max, dt_min, gravity, ratio=0.5):
        self.osi = osi
        self.dt_max = dt_max
        self.dt_min = dt_min
        self.gravity = gravity
        self.ratio = ratio

        self._parameters = [self.op_type, self.dt_max, self.dt_min, self.gravity, self.ratio]
        self.to_process(osi)



from o3seespy.base_model import OpenSeesObject


class SystemBase(OpenSeesObject):
    op_base_type = "system"


class BandGen(SystemBase):
    """
    The BandGen System Class
    
    This command is used to construct a BandGeneralSOE linear system of equation object. As the name implies, this class
    is used for matrix systems which have a banded profile. The matrix is stored as shown below in a 1dimensional array of
    size equal to the bandwidth times the number of unknowns. When a solution is required, the Lapack routines DGBSV and
    SGBTRS are used.
    """
    op_type = 'BandGen'

    def __init__(self, osi):
        """
        Initial method for BandGen

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.system.BandGen(osi)
        """
        self.osi = osi
        self._parameters = [self.op_type]
        self.to_process(osi)


class BandSPD(SystemBase):
    """
    The BandSPD System Class
    
    This command is used to construct a BandSPDSOE linear system of equation object. As the name implies, this class is
    used for symmetric positive definite matrix systems which have a banded profile. The matrix is stored as shown below
    in a 1 dimensional array of size equal to the (bandwidth/2) times the number of unknowns. When a solution is
    required, the Lapack routines DPBSV and DPBTRS are used.
    """
    op_type = 'BandSPD'

    def __init__(self, osi):
        """
        Initial method for BandSPD

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.system.BandSPD(osi)
        """
        self.osi = osi
        self._parameters = [self.op_type]
        self.to_process(osi)


class ProfileSPD(SystemBase):
    """
    The ProfileSPD System Class
    
    This command is used to construct a profileSPDSOE linear system of equation object. As the name implies, this class
    is used for symmetric positive definite matrix systems. The matrix is stored as shown below in a 1 dimensional array
    with only those values below the first non-zero row in any column being stored. This is sometimes also referred to
    as a skyline storage scheme.
    """
    op_type = 'ProfileSPD'

    def __init__(self, osi):
        """
        Initial method for ProfileSPD

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.system.ProfileSPD(osi)
        """
        self.osi = osi
        self._parameters = [self.op_type]
        self.to_process(osi)


class ParallelProfileSPD(SystemBase):
    """
    The ParallelProfileSPD System Class

    This command is used to construct a parallel version of the profileSPDSOE linear system of equation object.
    As the name implies, this class
    is used for symmetric positive definite matrix systems. The matrix is stored as shown below in a 1 dimensional array
    with only those values below the first non-zero row in any column being stored. This is sometimes also referred to
    as a skyline storage scheme.
    """
    op_type = 'ParallelProfileSPD'

    def __init__(self, osi):
        """
        Initial method for ParallelProfileSPD

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.system.ParallelProfileSPD(osi)
        """
        self.osi = osi
        self._parameters = [self.op_type]
        self.to_process(osi)


class SuperLU(SystemBase):
    """
    The SuperLU System Class
    
    This command is used to construct a SparseGEN linear system of equation object. As the name implies, this class is
    used for sparse matrix systems. The solution of the sparse matrix is carried out using `SuperLU`_.
    """
    op_type = 'SuperLU'

    def __init__(self, osi):
        """
        Initial method for SuperLU

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.system.SuperLU(osi)
        """
        self.osi = osi
        self._parameters = [self.op_type]
        self.to_process(osi)


class UmfPack(SystemBase):
    """
    The UmfPack System Class
    
    This command is used to construct a sparse system of equations which uses the `UmfPack`_ solver.
    """
    op_type = 'UmfPack'

    def __init__(self, osi):
        """
        Initial method for UmfPack

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.system.UmfPack(osi)
        """
        self.osi = osi
        self._parameters = [self.op_type]
        self.to_process(osi)


class FullGeneral(SystemBase):
    """
    The FullGeneral System Class
    
    This command is used to construct a Full General linear system of equation object. As the name implies, the class
    utilizes NO space saving techniques to cut down on the amount of memory used. If the matrix is of size, nxn, then
    storage for an nxn array is sought from memory when the program runs. When a solution is required, the Lapack
    routines DGESV and DGETRS are used... note::This type of system should almost never be used! This is because
    it requires a lot more memory than every other solver and takes more time in the actal solving operation
    than any other solver. It is required if the user is interested in looking at the global system matrix.
    """
    op_type = 'FullGeneral'

    def __init__(self, osi):
        """
        Initial method for FullGeneral

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.system.FullGeneral(osi)
        """
        self.osi = osi
        self._parameters = [self.op_type]
        self.to_process(osi)


class SparseSYM(SystemBase):
    """
    The SparseSYM System Class
    
    This command is used to construct a sparse symmetric system of equations which uses a row-oriented solution method
    in the solution phase.
    """
    op_type = 'SparseSYM'

    def __init__(self, osi):
        """
        Initial method for SparseSYM

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance

        Examples
        --------
        >>> import o3seespy as o3
        >>> # Example is currently not working
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.system.SparseSYM(osi)
        """
        self.osi = osi
        self._parameters = [self.op_type]
        self.to_process(osi)


class Mumps(SystemBase):
    """
    The Mumps System Class
    
    Create a system of equations using the Mumps solver
    """
    op_type = 'Mumps'

    def __init__(self, osi, icntl14=None, icntl7=None):
        """
        Initial method for Mumps

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        icntl14: None, optional
            
        icntl7: None, optional
            
        """
        self.osi = osi
        self.icntl14 = icntl14
        self.icntl7 = icntl7
        self._parameters = [self.op_type]
        if getattr(self, 'icntl14') is not None:
            self._parameters += ['-ICNTL14', self.icntl14]
        if getattr(self, 'icntl7') is not None:
            self._parameters += ['-ICNTL7', self.icntl7]
        self.to_process(osi)


class BandGeneral(SystemBase):
    op_type = "BandGeneral"

    def __init__(self, osi):
        self.osi = osi
        self._parameters = [self.op_type]
        self.to_process(osi)


class ParallelBandGeneral(SystemBase):
    op_type = "ParallelBandGeneral"

    def __init__(self, osi):
        self.osi = osi
        self._parameters = [self.op_type]
        self.to_process(osi)


class SparseGeneral(SystemBase):
    op_type = "SparseGeneral"

    def __init__(self, osi):
        self.osi = osi
        self._parameters = [self.op_type]
        self.to_process(osi)


def apply_mumps_or_sparse_general(osi, **kwargs):
    if osi.mp:
        Mumps(osi, **kwargs)
    else:
        SparseGeneral(osi)





def build_graph_links(mesh_eles, inactive_value=-1):
    import numpy as np
    mesh_eles = np.where(mesh_eles == inactive_value, -1, mesh_eles)
    mflat = mesh_eles.flatten()
    mflat_cleaned = mflat[np.where(mflat > -1)]
    graph_inds = np.arange(len(mflat_cleaned), dtype=int)
    mlist = np.ones(np.max(mesh_eles) + 1, dtype=int) * (len(mflat_cleaned) + 2)  # sets Null values to exceed length
    mlist[mflat_cleaned] = graph_inds
    rolled_down = np.roll(mesh_eles, 1, axis=0)
    rolled_down[0, :] = -1
    rolled_up = np.roll(mesh_eles, -1, axis=0)
    rolled_up[-1, :] = -1
    rolled_left = np.roll(mesh_eles, -1, axis=1)
    rolled_left[:, -1] = -1
    rolled_right = np.roll(mesh_eles, 1, axis=1)
    rolled_right[:, 0] = -1
    rolled_down_right = np.roll(rolled_down, 1, axis=1)
    rolled_down_right[:, 0] = -1
    rolled_down_left = np.roll(rolled_down, -1, axis=1)
    rolled_down_left[:, -1] = -1
    rolled_up_right = np.roll(rolled_up, 1, axis=1)
    rolled_up_right[:, 0] = -1
    rolled_up_left = np.roll(rolled_up, -1, axis=1)
    rolled_up_left[:, -1] = -1
    all_rolled = np.array([rolled_down, rolled_up, rolled_left, rolled_right,
                           rolled_down_left, rolled_down_right, rolled_up_left, rolled_up_right])

    g_adjs = []
    for i in range(len(mesh_eles)):
        for j in range(len(mesh_eles[i])):
            all_inds = all_rolled[:, i, j]
            inds = all_inds[np.where(all_inds > -1)]
            g_adjs.append((mlist[inds]))

    return g_adjs, mlist


def build_nodes_in_partition_2dmatrix(ele_partitions, pid):
    import numpy as np
    partitions = np.pad(ele_partitions, ((0, 1), (0, 1)), 'constant', constant_values=-1)
    ni = len(partitions)
    nj = len(partitions[0])
    rolled_down = np.roll(partitions, 1, axis=0)
    rolled_right = np.roll(partitions, 1, axis=1)
    rolled_down_right = np.roll(rolled_down, 1, axis=1)
    all_rolled = np.array([partitions, rolled_down, rolled_right, rolled_down_right]).transpose([1, 2, 0])
    all_rolled_f = all_rolled.reshape(ni * nj, 4)
    act = np.any(np.isin(all_rolled_f, [pid]), axis=1)
    act = act.astype(dtype=np.int32)
    act = act.reshape([ni, nj])
    return act


def is_node_in_partion(ele_partitions, node_i, node_j, pid):
    return pid in ele_partitions[node_i:min(len(ele_partitions), node_i + 1)][node_j: min(len(ele_partitions[0]), node_j)]



from o3seespy.command.element.base_element import ElementBase


class ZeroLength(ElementBase):
    """
    The ZeroLength Element Class
    
    This command is used to construct a zeroLength element object, which is defined by two nodes at the same location.
    The nodes are connected by multiple UniaxialMaterial objects to represent the force-deformation relationship for the
    element.
    """
    op_type = 'zeroLength'

    def __init__(self, osi, ele_nodes, mats: list=None, dirs: list=None, r_flag: float=None, orient: list=None):
        """
        Initial method for ZeroLength

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of two element nodes
        mats: list, optional
            A list of objects associated with previously-defined uniaxial_materials
        dirs: list, optional
            A list of material directions: * 1,2,3 - translation along local x,y,z axes, respectively; * 4,5,6 -
            rotation about local x,y,z axes, respectively
        r_flag: float, optional
            Optional, default = 0 * ``rflag`` = 0 no rayleigh damping (default)
        orient: list, optional
            

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [0, 0]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(2)]
        >>> bilinear_mat = o3.uniaxial_material.Steel01(osi, fy=300., e0=200.0e3, b=0.01)
        >>> o3.element.ZeroLength(osi, ele_nodes, mats=[bilinear_mat], dirs=[o3.cc.DOF2D_X], r_flag=1)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        if mats is None:
            self.mats = None
        else:
            self.mats = [x.tag for x in mats]
        self.dirs = dirs
        if r_flag is None:
            self.r_flag = None
        else:
            self.r_flag = float(r_flag)
        self.orient = orient
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags]
        if getattr(self, 'mats') is not None:
            self._parameters += ['-mat', *self.mats]
        if getattr(self, 'dirs') is not None:
            self._parameters += ['-dir', *self.dirs]
        if getattr(self, 'r_flag') is not None:
            self._parameters += ['-doRayleigh', self.r_flag]
        if getattr(self, 'orient') is not None:
            self._parameters += ['--orient', *self.orient]
        self.to_process(osi)


class ZeroLengthND(ElementBase):
    """
    The ZeroLengthND Element Class
    
    This command is used to construct a zeroLengthND element object, which is defined by two nodes at the same location.
    The nodes are connected by a single NDMaterial object to represent the force-deformation relationship for the element.
    """
    op_type = 'zeroLengthND'

    def __init__(self, osi, ele_nodes, mat, uni, orient: list=None):
        """
        Initial method for ZeroLengthND

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of two element nodes
        mat: obj
            Object associated with previously-defined ndmaterial object
        uni: obj
            Object associated with previously-defined uniaxial_material object which may be used to represent uncoupled
            behavior orthogonal to the plane of the ndmaterial response. see notes 2 and 3.
        orient: list, optional
            

        Examples
        --------
        >>> import o3seespy as o3
        >>> # Example is currently not working
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [0, 0]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(2)]
        >>> mat = o3.nd_material.ElasticIsotropic(osi, 1.0, 0.3)
        >>> uni = o3.uniaxial_material.Elastic(osi, 1.0)
        >>> o3.element.ZeroLengthND(osi, ele_nodes=ele_nodes, mat=mat, uni=uni, orient=[1, 2, 3, 4, 5, 6])
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.mat = mat
        self.uni = uni
        self.orient = orient
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.mat.tag, self.uni.tag]
        if getattr(self, 'orient') is not None:
            self._parameters += ['--orient', *self.orient]
        self.to_process(osi)


class ZeroLengthSection(ElementBase):
    """
    The ZeroLengthSection Element Class
    
    This command is used to construct a zero length element object, which is defined by two nodes at the same location.
    The nodes are connected by a single section object to represent the force-deformation relationship for the element.
    """
    op_type = 'zeroLengthSection'

    def __init__(self, osi, ele_nodes, sec, r_flag: float=None, orient: list=None):
        """
        Initial method for ZeroLengthSection

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of two element nodes
        sec: obj
            Object associated with previously-defined section object
        r_flag: float, optional
            Optional, default = 0 * ``rflag`` = 0 no rayleigh damping (default) * ``rflag`` = 1 include rayleigh damping
        orient: list, optional
            

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [0, 0]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(2)]
        >>> sec = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> o3.element.ZeroLengthSection(osi, ele_nodes=ele_nodes, sec=sec, r_flag=1.0, orient=[1])
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.sec = sec
        if r_flag is None:
            self.r_flag = None
        else:
            self.r_flag = float(r_flag)
        self.orient = orient
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.sec.tag]
        if getattr(self, 'r_flag') is not None:
            self._parameters += ['-doRayleigh', self.r_flag]
        if getattr(self, 'orient') is not None:
            self._parameters += ['--orient', *self.orient]
        self.to_process(osi)


class CoupledZeroLength(ElementBase):
    """
    The CoupledZeroLength Element Class
    
    
    """
    op_type = 'CoupledZeroLength'

    def __init__(self, osi, ele_nodes, dirn1, dirn2, mat, r_flag=1):
        """
        Initial method for CoupledZeroLength

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of two element nodes
        dirn1: int
            The two directions, 1 through ndof.
        dirn2: int
            The two directions, 1 through ndof.
        mat: obj
            Objects associated with previously-defined uniaxial_material
        r_flag: float, optional
            Optional, default = 0 * ``rflag`` = 0 no rayleigh damping (default) * ``rflag`` = 1 include rayleigh damping

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [0, 0]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(2)]
        >>> mat = o3.uniaxial_material.Elastic(osi, 1.0)
        >>> o3.element.CoupledZeroLength(osi, ele_nodes=ele_nodes, dirn1=1, dirn2=1, mat=mat, r_flag=1)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.dirn1 = int(dirn1)
        self.dirn2 = int(dirn2)
        self.mat = mat
        self.r_flag = float(r_flag)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.dirn1, self.dirn2, self.mat.tag, self.r_flag]
        self.to_process(osi)


class ZeroLengthContact2Dnormal(ElementBase):
    """
    The ZeroLengthContact2Dnormal Element Class
    
    This command is used to construct a zeroLengthContact2D element, which is Node-to-node frictional contact element
    used in two dimensional analysis and three dimensional analysis:
    """
    op_type = 'zeroLengthContact2D'

    def __init__(self, osi, ele_nodes, kn, kt, mu, nx, ny):
        """
        Initial method for ZeroLengthContact2Dnormal

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of a constrained and a retained nodes
        kn: float
            Penalty in normal direction
        kt: float
            Penalty in tangential direction
        mu: float
            Friction coefficient
        nx: None
            
        ny: None
            

        Examples
        --------
        >>> import o3seespy as o3
        >>> # Example is currently not working
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [0, 0]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(2)]
        >>> o3.element.ZeroLengthContact2Dnormal(osi, ele_nodes=ele_nodes, kn=1.0, kt=1.0, mu=1.0, nx=1, ny=0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.kn = float(kn)
        self.kt = float(kt)
        self.mu = float(mu)
        self.nx = nx
        self.ny = ny
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.kn, self.kt, self.mu, '-normal', self.nx, self.ny]
        self.to_process(osi)

class ZeroLengthContact3D(ElementBase):
    """
    The ZeroLengthContact3D Element Class
    
    This command is used to construct a zeroLengthContact3D element, which is Node-to-node frictional contact element
    used in two dimensional analysis and three dimensional analysis:
    """
    op_type = 'zeroLengthContact3D'

    def __init__(self, osi, ele_nodes, kn, kt, mu, c, dir):
        """
        Initial method for ZeroLengthContact3D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of a constrained and a retained nodes
        kn: float
            Penalty in normal direction
        kt: float
            Penalty in tangential direction
        mu: float
            Friction coefficient
        c: float
            Cohesion (not available in 2d)
        dir: int
            Direction flag of the contact plane (3d), it can be: * 1 out normal of the master plane pointing to +x
            direction * 2 out normal of the master plane pointing to +y direction * 3 out normal of the master plane pointing
            to +z direction

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(2)]
        >>> o3.element.ZeroLengthContact3D(osi, ele_nodes=ele_nodes, kn=1.0, kt=1.0, mu=1.0, c=1.0, dir=1)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.kn = float(kn)
        self.kt = float(kt)
        self.mu = float(mu)
        self.c = float(c)
        self.dir = int(dir)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.kn, self.kt, self.mu, self.c, self.dir]
        self.to_process(osi)


class ZeroLengthContactNTS2D(ElementBase):
    """
    The ZeroLengthContactNTS2D Element Class
    
    
    """
    op_type = 'zeroLengthContactNTS2D'

    def __init__(self, osi, kn, kt, phi, s_nd_num: int=None, m_nd_num: int=None, nodes: list=None):
        """
        Initial method for ZeroLengthContactNTS2D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        kn: float
            Penalty in normal direction
        kt: float
            Penalty in tangential direction
        phi: float
            Friction angle in degrees
        s_nd_num: int, optional
            Number of slave nodes
        m_nd_num: int, optional
            Number of master nodes
        nodes: list, optional
            Slave and master node objects respectively

        Examples
        --------
        >>> import o3seespy as o3
        >>> # Example is currently not working
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [0, 0]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(2)]
        >>> o3.element.ZeroLengthContactNTS2D(osi, s_nd_num=1, m_nd_num=1, nodes=ele_nodes, kn=1.0, kt=1.0, phi=1.0)
        """
        self.osi = osi
        if s_nd_num is None:
            self.s_nd_num = None
        else:
            self.s_nd_num = int(s_nd_num)
        if m_nd_num is None:
            self.m_nd_num = None
        else:
            self.m_nd_num = int(m_nd_num)
        if nodes is None:
            self.nodes = None
        else:
            self.nodes = [x.tag for x in nodes]
        self.kn = float(kn)
        self.kt = float(kt)
        self.phi = float(phi)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, self.kn, self.kt, self.phi]
        if getattr(self, 's_nd_num') is not None:
            self._parameters += ['-sNdNum', self.s_nd_num]
        if getattr(self, 'm_nd_num') is not None:
            self._parameters += ['-mNdNum', self.m_nd_num]
        if getattr(self, 'nodes') is not None:
            self._parameters += ['-Nodes', *self.nodes]
        self.to_process(osi)


class ZeroLengthInterface2Ddof(ElementBase):
    """
    The ZeroLengthInterface2Ddof Element Class
    
    
    """
    op_type = 'zeroLengthInterface2D'

    def __init__(self, osi, sdof, mdof, kn, kt, phi, s_nd_num: int=None, m_nd_num: int=None, nodes: list=None):
        """
        Initial method for ZeroLengthInterface2Ddof

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        sdof: int
            Slave and master degree of freedom
        mdof: int
            Slave and master degree of freedom
        kn: float
            Penalty in normal direction
        kt: float
            Penalty in tangential direction
        phi: float
            Friction angle in degrees
        s_nd_num: int, optional
            Number of slave nodes
        m_nd_num: int, optional
            Number of master nodes
        nodes: list, optional
            Slave and master node objects respectively

        Examples
        --------
        >>> import o3seespy as o3
        >>> # Example is currently not working
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> nodes = [1, 1]
        >>> o3.element.ZeroLengthInterface2Ddof(osi, s_nd_num=1, m_nd_num=1, sdof=1, mdof=1, nodes=nodes, kn=1.0, kt=1.0, phi=1.0)
        """
        self.osi = osi
        if s_nd_num is None:
            self.s_nd_num = None
        else:
            self.s_nd_num = int(s_nd_num)
        if m_nd_num is None:
            self.m_nd_num = None
        else:
            self.m_nd_num = int(m_nd_num)
        self.sdof = int(sdof)
        self.mdof = int(mdof)
        if nodes is None:
            self.nodes = None
        else:
            self.nodes = [x.tag for x in nodes]
        self.kn = float(kn)
        self.kt = float(kt)
        self.phi = float(phi)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, '-dof', self.sdof, self.mdof, self.kn, self.kt, self.phi]
        if getattr(self, 's_nd_num') is not None:
            self._parameters += ['-sNdNum', self.s_nd_num]
        if getattr(self, 'm_nd_num') is not None:
            self._parameters += ['-mNdNum', self.m_nd_num]
        if getattr(self, 'nodes') is not None:
            self._parameters += ['-Nodes', *self.nodes]
        self.to_process(osi)


class ZeroLengthImpact3D(ElementBase):
    """
    The ZeroLengthImpact3D Element Class
    
    This command constructs a node-to-node zero-length contact element in 3D space to simulate the impact/pounding and
    friction phenomena.
    """
    op_type = 'zeroLengthImpact3D'

    def __init__(self, osi, ele_nodes, direction, init_gap, friction_ratio, kt, kn, kn2, delta_y, cohesion):
        """
        Initial method for ZeroLengthImpact3D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of a constrained and a retained nodes * ``1`` if out-normal vector of master plane points to +x
            direction * ``2`` if out-normal vector of master plane points to +y direction * ``3`` if out-normal vector of
            master plane points to +z direction
        direction: None
            
        init_gap: float
            Initial gap between master plane and slave plane
        friction_ratio: float
            Friction ratio in two tangential directions (parallel to master and slave planes)
        kt: float
            Penalty in two tangential directions
        kn: float
            Penalty in normal direction (normal to master and slave planes)
        kn2: float
            Penalty in normal direction after yielding based on hertz impact model
        delta_y: float
            Yield deformation based on hertz impact model
        cohesion: float
            Cohesion, if no cohesion, it is zero

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
        >>> o3.element.ZeroLengthImpact3D(osi, ele_nodes=ele_nodes, direction=1, init_gap=1.0, friction_ratio=1.0, kt=1.0, kn=1.0, kn2=1.0, delta_y=1.0, cohesion=1.0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.direction = direction
        self.init_gap = float(init_gap)
        self.friction_ratio = float(friction_ratio)
        self.kt = float(kt)
        self.kn = float(kn)
        self.kn2 = float(kn2)
        self.delta_y = float(delta_y)
        self.cohesion = float(cohesion)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.direction, self.init_gap, self.friction_ratio, self.kt, self.kn, self.kn2, self.delta_y, self.cohesion]
        self.to_process(osi)



from collections import OrderedDict
try:
    from custom_openseespy import custom_opensees as opy
except ModuleNotFoundError:
    import openseespy.opensees as opy
from o3seespy import exceptions
from o3seespy import extensions


class OpenSeesObject(object):
    op_base_type = "<not-set>"  # used to call opensees module
    op_type = "<not-set>"  # string name given to object in opensees module
    _tag = None
    _name = None
    _parameters = None

    @property  # deliberately no setter method
    def tag(self):
        return self._tag

    @property
    def name(self):
        return self._name

    def to_process(self, osi):
        if osi.state == 0:
            return self.to_opensees()
        # elif osi.state == 1:
        #     osi.to_commands(self.to_commands())
        # elif osi.state == 2:
        #     osi.to_dict(self)
        elif osi.state == 3:
            osi.to_commands(self.to_commands())
            return self.to_opensees()
        elif osi.state == 4:
            osi.to_commands(self.to_commands())

    def to_opensees(self):
        try:
            try:
                return getattr(opy, self.op_base_type)(*self.parameters)
            except opy.OpenSeesError as e:
                com = extensions.to_commands(self.op_base_type, self.parameters)
                raise ValueError('{0} caused error "{1}"'.format(com, e))

        except SystemError as e:
            if None in self.parameters:
                print(self.parameters)
                raise exceptions.ModelError("%s of type: %s contains 'None'" % (self.op_base_type, self.op_type))
            else:
                raise SystemError(e)
        except AttributeError as e:
            print(e)
            print('opensees.{0}({1}) caused error "{2}"'.format(self.op_base_type,
                                                                               ','.join(str(x) for x in self.parameters),
                                                                                        e))
            raise exceptions.ModelError("op_base_type: '%s' does not exist in opensees module" % self.op_base_type)

    def to_commands(self):
        return extensions.to_commands(self.op_base_type, self.parameters)

    @property
    def parameters(self):
        return self._parameters

    @property
    def type(self):
        return self.op_type

    @property
    def base_type(self):
        return self.op_base_type

    def to_dict(self, export_none=False):
        outputs = OrderedDict()
        for item in self.__dict__:
            if '_' == item[0]:  # do not export private variables
                continue
            value = self.__getattribute__(item)
            if not export_none and value is None:
                continue
            outputs[item] = collect_serial_value(value)
        return outputs


class OpenSeesMultiCallObject(OpenSeesObject):
    _multi_parameters = None

    @property
    def parameters(self):
        return self._multi_parameters[-1]

    @property
    def multi_parameters(self):
        return self._multi_parameters


def collect_serial_value(value):
    if isinstance(value, str):
        return value
    elif isinstance(value, int):
        return value
    elif hasattr(value, "to_dict"):
        return value.to_dict()
    elif hasattr(value, "__len__"):
        tolist = getattr(value, "tolist", None)
        if callable(tolist):
            value = value.tolist()
            return value
        else:
            if hasattr(value, "tag"):
                return value.tag
            elif hasattr(value, "to_dict"):
                value = value.to_dict()
                return value
            else:
                values = []
                for item in value:
                    values.append(collect_serial_value(item))
                return values
    else:
        return value





from o3seespy.base_model import OpenSeesObject


class LayerBase(OpenSeesObject):
    op_base_type = "layer"


class Straight(LayerBase):
    """
    The Straight Layer Class
    
    The layer command is used to generate a number of fibers along a line or a circular arc.
    """
    op_type = 'straight'

    def __init__(self, osi, mat, num_fiber, area_fiber, start, end):
        """
        Initial method for Straight

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        mat: obj
            Material object associated with this fiber (uniaxial_material object for a fibersection and ndmaterial
            object for use in an ndfibersection).
        num_fiber: int
            Number of fibers along line
        area_fiber: float
            Area of each fiber
        start: list
            Y & z-coordinates of first fiber in line (local coordinate system)
        end: list
            Y & z-coordinates of last fiber in line (local coordinate system)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> start = [1.0, 1.0]
        >>> end = [1.0, 1.0]
        >>> rebar = o3.uniaxial_material.Steel01(osi, fy=60.0, e0=30000.0, b=0.02)
        >>> o3.layer.Straight(osi, rebar, num_fiber=1, area_fiber=1.0, start=start, end=end)
        """
        self.osi = osi
        self.mat = mat
        self.num_fiber = int(num_fiber)
        self.area_fiber = float(area_fiber)
        self.start = start
        self.end = end
        self._parameters = [self.op_type, self.mat.tag, self.num_fiber, self.area_fiber, *self.start, *self.end]
        self.to_process(osi)

class Circ(LayerBase):
    """
    The Circ Layer Class

    This command is used to construct a line of fibers along a circular arc
    """
    op_type = 'circ'

    def __init__(self, osi, mat, num_fiber, area_fiber, center, radius, ang=None):
        """
        Initial method for Circ

        Supports pre-building

        Parameters
        ----------
        mat: obj
            Material tag associated with this fiber (uniaxial_material for a fiber section and nd_material for use
            in an nd fiber section).
        num_fiber: int
            Number of fibers along line
        area_fiber: float
            Area of each fiber
        center: listf
            Y & z-coordinates of center of circular arc
        radius: float
            Radius of circlular arc
        ang: listf
            Starting and ending angle (optional) [0.0, 360.0-360/num_fibres]

        """
        self.osi = osi
        self.mat = mat
        self.num_fiber = int(num_fiber)
        self.area_fiber = float(area_fiber)
        self.center = center
        self.radius = float(radius)
        self.ang = ang
        self._parameters = [self.op_type, self.mat.tag, self.num_fiber, self.area_fiber, *self.center, self.radius]
        if self.ang is not None:
            self._parameters += self.ang
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)




import eqsig
from eqsig import sdof
import numpy as np

from o3seespy import opy
from o3seespy import cc as opc


def get_inelastic_response(mass, k_spring, f_yield, motion, dt, xi=0.05, r_post=0.0):
    """
    Run seismic analysis of a nonlinear SDOF

    :param mass: SDOF mass
    :param k_spring: spring stiffness
    :param f_yield: yield strength
    :param motion: list, acceleration values
    :param dt: float, time step of acceleration values
    :param xi: damping ratio
    :param r_post: post-yield stiffness
    :return:
    """

    opy.wipe()
    opy.model('basic', '-ndm', 2, '-ndf', 3)  # 2 dimensions, 3 dof per node

    # Establish nodes
    bot_node = 1
    top_node = 2
    opy.node(bot_node, 0., 0.)
    opy.node(top_node, 0., 0.)

    # Fix bottom node
    opy.fix(top_node, opc.FREE, opc.FIXED, opc.FIXED)
    opy.fix(bot_node, opc.FIXED, opc.FIXED, opc.FIXED)
    # Set out-of-plane DOFs to be slaved
    opy.equalDOF(1, 2, *[2, 3])

    # nodal mass (weight / g):
    opy.mass(top_node, mass, 0., 0.)

    # Define material
    bilinear_mat_tag = 1
    mat_type = "Steel01"
    mat_props = [f_yield, k_spring, r_post]
    opy.uniaxialMaterial(mat_type, bilinear_mat_tag, *mat_props)

    # Assign zero length element
    beam_tag = 1
    opy.element('zeroLength', beam_tag, bot_node, top_node, "-mat", bilinear_mat_tag, "-dir", 1, '-doRayleigh', 1)

    # Define the dynamic analysis
    load_tag_dynamic = 1
    pattern_tag_dynamic = 1

    values = list(-1 * motion)  # should be negative
    opy.timeSeries('Path', load_tag_dynamic, '-dt', dt, '-values', *values)
    opy.pattern('UniformExcitation', pattern_tag_dynamic, opc.DOF_X, '-accel', load_tag_dynamic)

    # set damping based on first eigen mode
    angular_freq2 = opy.eigen('-fullGenLapack', 1)
    if hasattr(angular_freq2, '__len__'):
        angular_freq2 = angular_freq2[0]
    angular_freq = angular_freq2 ** 0.5
    alpha_m = 0.0
    beta_k = 2 * xi / angular_freq
    beta_k_comm = 0.0
    beta_k_init = 0.0

    opy.rayleigh(alpha_m, beta_k, beta_k_init, beta_k_comm)

    # Run the dynamic analysis

    opy.wipeAnalysis()

    opy.algorithm('Newton')
    opy.system('SparseGeneral')
    opy.numberer('RCM')
    opy.constraints('Transformation')
    opy.integrator('Newmark', 0.5, 0.25)
    opy.analysis('Transient')

    tol = 1.0e-10
    iterations = 10
    opy.test('EnergyIncr', tol, iterations, 0, 2)
    analysis_time = (len(values) - 1) * dt
    analysis_dt = 0.001
    outputs = {
        "time": [],
        "rel_disp": [],
        "rel_accel": [],
        "rel_vel": [],
        "force": []
    }

    while opy.getTime() < analysis_time:
        curr_time = opy.getTime()
        opy.analyze(1, analysis_dt)
        outputs["time"].append(curr_time)
        outputs["rel_disp"].append(opy.nodeDisp(top_node, 1))
        outputs["rel_vel"].append(opy.nodeVel(top_node, 1))
        outputs["rel_accel"].append(opy.nodeAccel(top_node, 1))
        opy.reactions()
        outputs["force"].append(-opy.nodeReaction(bot_node, 1))  # Negative since diff node
    opy.wipe()
    for item in outputs:
        outputs[item] = np.array(outputs[item])

    return outputs


def test_sdof():
    """
    Create a plot of an elastic analysis, nonlinear analysis and closed form elastic

    :return:
    """

    from tests.conftest import TEST_DATA_DIR

    record_path = TEST_DATA_DIR
    record_filename = 'test_motion_dt0p01.txt'
    dt = 0.01
    rec = np.loadtxt(record_path + record_filename)
    acc_signal = eqsig.AccSignal(rec, dt)
    period = 1.0
    xi = 0.05
    mass = 1.0
    f_yield = 1.5  # Reduce this to make it nonlinear
    r_post = 0.0

    periods = np.array([period])
    resp_u, resp_v, resp_a = sdof.response_series(motion=rec, dt=dt, periods=periods, xi=xi)

    k_spring = 4 * np.pi ** 2 * mass / period ** 2
    outputs = get_inelastic_response(mass, k_spring, f_yield, rec, dt, xi=xi, r_post=r_post)
    outputs_elastic = get_inelastic_response(mass, k_spring, f_yield * 100, rec, dt, xi=xi, r_post=r_post)
    ux_opensees = outputs["rel_disp"]
    disp_inelastic_final = ux_opensees[-1]

    time = acc_signal.time
    acc_opensees_elastic = np.interp(time, outputs_elastic["time"], outputs_elastic["rel_accel"]) - rec
    ux_opensees_elastic = np.interp(time, outputs_elastic["time"], outputs_elastic["rel_disp"])
    diff_disp = abs(np.sum(ux_opensees_elastic - resp_u[0]))
    diff_acc = abs(np.sum(acc_opensees_elastic - resp_a[0]))
    assert diff_disp < 1.0e-4
    assert diff_acc < 5.0e-4
    assert np.isclose(disp_inelastic_final, 0.0186556)


if __name__ == '__main__':
    test_sdof()



try:
    from custom_openseespy import custom_opensees as opy
except ModuleNotFoundError:
    import openseespy.opensees as opy
from collections import OrderedDict
from o3seespy import exceptions, extensions


class OpenSeesInstance(object):

    def __init__(self, ndm: int, ndf=None, state=0, mp=False, nnpp=10000):
        init_tag = 0
        if mp:
            pid = opy.getPID()
            init_tag = pid * nnpp
        self.mp = mp
        self.n_node = init_tag
        self.n_con = init_tag
        self.n_ele = init_tag
        self.n_mat = init_tag
        self.n_sect = init_tag
        self.n_tseries = init_tag
        self.n_pat = init_tag
        self.n_fix = init_tag
        self.n_integ = init_tag
        self.n_transformation = init_tag
        self.n_region = init_tag
        self.n_params = init_tag
        self.n_mesh = init_tag
        self.ndm = ndm
        self._state = state  # 0=execute line by line, 1=export to raw openseespy, 2=export reloadable json
        parameters = ['BasicBuilder', '-ndm', ndm]
        if ndf is not None:
            if ndf not in [1, 2, 3, 6]:
                raise ValueError('ndm must be: 1, 2, 3, 6')
            self.ndf = int(ndf)
            parameters += ['-ndf', self.ndf]
        else:
            if ndm == 1:
                self.ndf = 1
            elif ndm == 2:
                self.ndf = 3
            else:
                self.ndf = 6
        opy.wipe()
        opy.model(*parameters)
        self.commands = []
        self.dict = OrderedDict()

        if state == 1:
            self.commands.append('opy.wipe()')
            self.commands.append(f"opy.model('basic', '-ndm', {ndm}, '-ndf', {self.ndf})")
        if state == 2:
            self.dict['ndm'] = ndm
            self.dict['ndf'] = ndf
            # base_types = ['node', 'element', 'section', 'uniaxial_material']
        elif state == 3:
            self.commands.append('opy.wipe()')
            self.commands.append(f"opy.model('basic', '-ndm', {ndm}, '-ndf', {self.ndf})")

    def reset_model_params(self, ndm, ndf):
        opy.model('BasicBuilder', '-ndm', ndm, '-ndf', ndf)
        self.ndm = ndm
        self.ndf = ndf
        if self._state == 1:
            self.commands.append(f"opy.model('basic', '-ndm', {ndm}, '-ndf', {ndf})")
        if self._state == 2:
            self.dict['ndm'] = ndm
            self.dict['ndf'] = ndf
            # base_types = ['node', 'element', 'section', 'uniaxial_material']
        elif self._state == 3:
            self.commands.append(f"opy.model('basic', '-ndm', {ndm}, '-ndf', {ndf})")

    def to_commands(self, os_command):
        self.commands.append(os_command)

    def to_dict(self, os_model, export_none=False):
        if os_model.op_type not in self.dict:
            self.dict[os_model.op_type] = OrderedDict()
        self.dict[os_model.op_type][os_model.tag] = os_model.to_dict(export_none=export_none)

    def to_process(self, op_base_type, parameters):
        if self.state == 0:
            return self.to_opensees(op_base_type, parameters)
        # if self.state == 1:
        #     self.to_commands(extensions.to_commands(op_base_type, parameters))
        # elif self.state == 2:
        #     self.to_dict(self)
        #     return self.to_opensees(op_base_type, parameters)
        elif self.state == 3:
            self.to_commands(extensions.to_commands(op_base_type, parameters))
            return self.to_opensees(op_base_type, parameters)

    def to_opensees(self, op_base_type, parameters):
        try:
            try:
                return getattr(opy, op_base_type)(*parameters)
            except opy.OpenSeesError as e:
                raise ValueError('opensees.{0}({1}) caused error "{2}"'.format(op_base_type,
                                                                               ','.join(str(x) for x in parameters), e))

        except SystemError as e:
            if None in parameters:
                print(parameters)
                raise exceptions.ModelError("%s of type: %s contains 'None'" % (op_base_type))
            else:
                raise SystemError(e)
        except AttributeError as e:
            print(e)
            print('opensees.{0}({1}) caused error "{2}"'.format(op_base_type,
                                                                               ','.join(str(x) for x in parameters),
                                                                                        e))
            raise exceptions.ModelError("op_base_type: '%s' does not exist in opensees module" % op_base_type)

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, value):
        self._state = value


class OpenseesInstance(OpenSeesInstance):
    def __init__(self, ndm: int, ndf=None, state=0):
        print('Please use OpenSeesInstance instead of OpenseesInstance')
        super(OpenseesInstance, self).__init__(ndm, ndf, state)


class _OpenSeesInstanceTestMP(OpenSeesInstance):
    def __init__(self, ndm: int, ndf=None, state=0, mp=False, nnpp=10000, pid=None):
        init_tag = 0
        if mp:
            if pid is None:
                pid = opy.getPID()
            init_tag = pid * nnpp
        self.mp = mp
        self.n_node = init_tag
        self.n_con = init_tag
        self.n_ele = init_tag
        self.n_mat = init_tag
        self.n_sect = init_tag
        self.n_tseries = init_tag
        self.n_pat = init_tag
        self.n_fix = init_tag
        self.n_integ = init_tag
        self.n_transformation = init_tag
        self.n_region = init_tag
        self.n_params = init_tag
        self.n_mesh = init_tag
        self.ndm = ndm
        self._state = state  # 0=execute line by line, 1=export to raw openseespy, 2=export reloadable json
        parameters = ['BasicBuilder', '-ndm', ndm]
        if ndf is not None:
            if ndf not in [1, 2, 3, 6]:
                raise ValueError('ndm must be: 1, 2, 3, 6')
            self.ndf = int(ndf)
            parameters += ['-ndf', self.ndf]
        else:
            if ndm == 1:
                self.ndf = 1
            elif ndm == 2:
                self.ndf = 3
            else:
                self.ndf = 6
        opy.wipe()
        opy.model(*parameters)
        self.commands = []
        self.dict = OrderedDict()

        if state == 1:
            self.commands.append('opy.wipe()')
            self.commands.append(f"opy.model('basic', '-ndm', {ndm}, '-ndf', {self.ndf})")
        if state == 2:
            self.dict['ndm'] = ndm
            self.dict['ndf'] = ndf
            # base_types = ['node', 'element', 'section', 'uniaxial_material']
        elif state == 3:
            self.commands.append('opy.wipe()')
            self.commands.append(f"opy.model('basic', '-ndm', {ndm}, '-ndf', {self.ndf})")
